import re
from typing import Iterator, List

import requests

from yival.data.base_reader import BaseReader
from yival.schemas.experiment_config import InputData
from yival.schemas.reader_configs import HuggingFaceDatasetReaderConfig


class HuggingFaceDatasetReader(BaseReader):
    config: HuggingFaceDatasetReaderConfig
    default_config = HuggingFaceDatasetReaderConfig(
        chunk_size=10000, example_limit=2, output_mapping={}
    )

    def __init__(self, config: HuggingFaceDatasetReaderConfig):
        super().__init__(config)
        self.config = config

    def read(self, path: str) -> Iterator[List[InputData]]:
        offset = 0
        limit = 100
        examples = []

        # Keep fetching until we reach the example_limit
        while offset < self.config.example_limit:
            path = f"{path}&offset={offset}&limit={limit}"
            response = requests.get(path)
            data = response.json()
            for item in data['rows']:
                transformed_item = {
                    self.config.output_mapping[key]: item['row'][key]
                    for key in self.config.output_mapping.keys()
                    if key in item['row']
                }
                if self.config.include:
                    if not any(
                        re.search(pattern, str(transformed_item.values()))
                        for pattern in self.config.include
                    ):
                        continue

                if self.config.exclude:
                    if any(
                        re.search(pattern, str(transformed_item.values()))
                        for pattern in self.config.exclude
                    ):
                        continue

                examples.append(InputData(content=transformed_item))

                # Break if we've reached the example_limit
                if len(examples) >= self.config.example_limit:
                    break

            offset += limit

            # If we've reached the example_limit, break out of the loop
            if len(examples) >= self.config.example_limit:
                break

        yield examples[:self.config.example_limit]


# Register the HuggingFaceDatasetReader with BaseReader again
BaseReader.register_reader(
    "huggingface_dataset_reader", HuggingFaceDatasetReader,
    HuggingFaceDatasetReaderConfig
)


def main():
    reader = HuggingFaceDatasetReader(
        HuggingFaceDatasetReaderConfig(
            chunk_size=1000,
            example_limit=100,
            output_mapping={'question': 'leetcode_problem'},
            include=['^(?!.*# Hard).*$']
        )
    )
    res = reader.read(
        "https://datasets-server.huggingface.co/rows?dataset=BoyuanJackchen%2Fleetcode_free_questions_text&config=default&split=train"
    )
    for i in res:
        print(i)


if __name__ == "__main__":
    main()



"""
OpenAIPromptBasedEvaluator is an evaluator that uses OpenAI's prompt-based
system for evaluations.

The evaluator interfaces with the OpenAI API to present tasks and interpret
the model's responses to determine the quality or correctness of a given
experiment result.
"""
import copy
import logging
import string
from typing import Any, Dict, Iterable, List, Optional, Union

# for exponential backoff
import openai
from tenacity import before_sleep_log, retry, stop_after_attempt, wait_random

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

from ..schemas.evaluator_config import (
    EvaluatorOutput,
    EvaluatorType,
    MethodCalculationMethod,
    MetricCalculatorConfig,
    OpenAIPromptBasedEvaluatorConfig,
)
from ..schemas.experiment_config import ExperimentResult, InputData, MultimodalOutput
from .base_evaluator import BaseEvaluator

CLASSIFY_STR = """
First, write out in a step by step manner your reasoning to be sure that your
conclusion is correct.
Avoid simply stating the correct answer at the outset.
Then print only a single choice from {choices} (without quotes or punctuation)
on its own line corresponding to the correct answer.
At the end, repeat just the answer by itself on a new line.
Reasoning:
"""

MATCH_FNS = {
    "include": lambda x, y: float(x in y),
    "exact": lambda x, y: float(x == y),
    "endswith": lambda x, y: x.endswith(y),
    "starts_or_endswith": lambda x, y: x.startswith(y) or x.endswith(y),
}


def extract_choice_from_response(
    response: str, choice_strings: Iterable[str]
) -> str:
    """Extracts the choice from the response string."""
    lines = response.strip().split("\n")
    for line in lines:
        sanitized_line = "".join(
            c for c in line if c not in string.punctuation
        ).strip()
        if not sanitized_line:
            continue
        for choice in choice_strings:
            if MATCH_FNS["exact"](sanitized_line, choice):
                return choice
    return "invalid response"


def calculate_choice_score(
    choice: str,
    choice_scores: Optional[Dict[str, float]] = None
) -> Optional[float]:
    """Calculates the score for the given choice."""
    if choice_scores is None:
        return None
    if choice == "invalid response":
        return min(choice_scores.values())
    return choice_scores.get(choice)


def format_template(
    template: Union[str, List[Dict[str, str]]], content: Dict[str, Any]
) -> Union[str, List[Dict[str, str]]]:
    """Formats a string or list template with the provided content."""
    if isinstance(template, str):
        try:
            return template.format(**content)
        except KeyError as e:
            raise ValueError(f"Missing key {e} in content dictionary")

    res = []
    for t in template:
        formatted_msg = copy.deepcopy(t)
        try:
            if "content" in formatted_msg:
                formatted_msg["content"] = formatted_msg['content'].format(
                    **content
                )
        except KeyError as e:
            raise ValueError(f"Missing key {e} in content dictionary")
        res.append(formatted_msg)
    return res


@retry(
    wait=wait_random(min=1, max=20),
    stop=stop_after_attempt(100),
    before_sleep=before_sleep_log(logger, logging.DEBUG)
)
def completion_with_backpff(**kwargs):
    response = openai.ChatCompletion.create(**kwargs)
    return response


def choices_to_string(choice_strings: Iterable[str]) -> str:
    """Converts a list of choices into a formatted string."""
    return " or ".join(f'"{choice}"' for choice in choice_strings)


class OpenAIPromptBasedEvaluator(BaseEvaluator):
    """Evaluator using OpenAI's prompt-based evaluation."""

    default_config = OpenAIPromptBasedEvaluatorConfig(
        name="openai_prompt_based_evaluator"
    )

    def __init__(self, config: OpenAIPromptBasedEvaluatorConfig):
        super().__init__(config)
        self.config = config

    def evaluate(self, experiment_result: ExperimentResult) -> EvaluatorOutput:
        """Evaluate the experiment result using OpenAI's prompt-based evaluation."""
        assert isinstance(self.config, OpenAIPromptBasedEvaluatorConfig)
        format_dict = copy.deepcopy(experiment_result.input_data.content)
        format_dict["raw_output"] = experiment_result.raw_output.text_output

        prompt = format_template(self.config.prompt, format_dict)
        if isinstance(prompt, str):
            prompt = [{"role": "user", "content": prompt}]

        prompt[-1]["content"] += "\n\n" + CLASSIFY_STR.format(
            choices=choices_to_string(self.config.choices)
        )
        response = completion_with_backpff(
            model="gpt-4",
            messages=prompt,
            temperature=0.5,
            n=1,
            max_tokens=1000,
            request_timeout=60,
        )
        #response = openai.ChatCompletion.create(model="gpt-4", messages=prompt, temperature=0.5)
        response_content = response['choices'][0]['message']['content']
        choice = extract_choice_from_response(
            response_content, self.config.choices
        )
        score = calculate_choice_score(choice, self.config.choice_scores)
        return EvaluatorOutput(
            name=self.config.name,
            result=score if score is not None else choice,
            display_name=self.config.display_name,
            metric_calculators=self.config.metric_calculators
        )


BaseEvaluator.register_evaluator(
    "openai_prompt_based_evaluator", OpenAIPromptBasedEvaluator,
    OpenAIPromptBasedEvaluatorConfig
)


def main():
    """Main function to test the OpenAIPromptBasedEvaluator."""
    evaluator_config = OpenAIPromptBasedEvaluatorConfig(
        name="openai_prompt_based_evaluator",
        display_name="math calculator",
        metric_calculators=[
            MetricCalculatorConfig(
                MethodCalculationMethod(MethodCalculationMethod.AVERAGE)
            )
        ],
        prompt="{problem}\n\n Is the answer '{raw_output}' correct? .",
        choices=["Yes", "No"],
        evaluator_type=EvaluatorType.INDIVIDUAL,
        choice_scores={
            "Yes": 1.0,
            "No": 0
        },
    )
    input_data_example = InputData(
        content={
            "problem": "Calculate the area of a circle with radius 5.",
            "method": "Using the formula for the area of a circle: pi*r^2"
        }
    )

    experiment_result_example = ExperimentResult(
        input_data=input_data_example,
        combination={
            "wrapper1": "var1",
            "wrapper2": "var2"
        },
        raw_output=MultimodalOutput(
            text_output="The area of the circle is 78.54 square units."
        ),
        latency=150.0,
        token_usage=50
    )

    evaluator = OpenAIPromptBasedEvaluator(evaluator_config)
    result = evaluator.evaluate(experiment_result_example)
    print(result)


if __name__ == "__main__":
    main()



from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional, Union

from .model_configs import CallOption


@dataclass
class BaseDataGeneratorConfig:
    """
    Base configuration class for all data generators.
    """

    chunk_size: int = 1000
    number_of_examples: int = 10
    output_path: Optional[str] = None

    def asdict(self):
        return asdict(self)


@dataclass
class OpenAIPromptBasedGeneratorConfig(BaseDataGeneratorConfig):
    """
    Generate test cases from prompt. Currently only support openai models.
    """
    model_name: str = "gpt-4"
    prompt: Union[str, List[Dict[str, str]]] = """
        "Please provide a concrete and realistic test case as a dictionary for function invocation using the ** operator.
        Only include parameters, excluding description and name.
        Ensure it's succinct and well-structured.
        **Only provide the dictionary.**"    
    """
    input_function: Dict[str, Any] = field(default_factory=dict)
    # Whether to diversify the generated examples.
    diversify: bool = True
    max_token = 2000

    # Expected Value name
    expected_param_name: str = ""

    # Llm call option
    call_option: Optional[CallOption] = None

    output_csv_path: Optional[str] = None

    fixed_input: Optional[Dict[str, Any]] = None

    def asdict(self):
        return asdict(self)



"""
Common util funtions.
"""
import asyncio
import json
import os
import time
from collections import deque

import aiohttp
import openai
from aiohttp_socks import ProxyConnector  # type: ignore

SECONDS_TO_PAUSE_AFTER_RATE_LIMIT_ERROR = 15
MAX_REQUESTS_PER_MINUTE = 100
MAX_TOKENS_PER_MINUTE = 35000


class RateLimiter:
    """
    A rate limiter that ensures requests don't exceed a given rate and token
    usage.

    The rate limiter checks two conditions:
    1. The rate at which requests are being made.
    2. The number of tokens being used per minute.

    Attributes:
        max_rate (int): Maximum number of requests allowed per second.
        max_tokens_per_minute (int): Maximum number of tokens allowed to be
                                    used per minute.
        start_time (float): The start time when the rate limiter was
                            initialized.
        request_count (int): Number of requests made since the rate limiter
                            was initialized.
        token_usage (deque): A deque containing tuples of tokens used and the
                            time they were used at.
    """

    def __init__(self, max_rate, max_tokens_per_minute):
        """
        Initialize the rate limiter with maximum rate and tokens per minute.

        Args:
            max_rate (int): Maximum number of requests allowed per second.
            max_tokens_per_minute (int): Maximum number of tokens allowed to
                                        be used per minute.
        """
        self.max_rate = max_rate
        self.max_tokens_per_minute = max_tokens_per_minute
        self.start_time = time.time()
        self.request_count = 0
        self.token_usage = deque()

    async def wait(self):
        """
        Wait until it's safe to make another request based on the rate and
        token limits.

        This method calculates the expected time for the next request and
        waits if necessary.
        It also ensures that the token usage does not exceed the specified
        limit.
        """
        self.request_count += 1
        elapsed_time = time.time() - self.start_time
        expected_time = self.request_count / self.max_rate
        sleep_time = max(expected_time - elapsed_time, 0)
        await asyncio.sleep(sleep_time)

        # Check if tokens used more than a minute ago, and remove them
        while self.token_usage and self.token_usage[0][1] < time.time() - 60:
            self.token_usage.popleft()

        # Check if the current token count exceeds the limit
        current_tokens = sum(token for token, _ in self.token_usage)
        while current_tokens >= self.max_tokens_per_minute:
            await asyncio.sleep(1)  # Wait a second and check again
            while self.token_usage and self.token_usage[0][1] < time.time(
            ) - 60:
                self.token_usage.popleft()
            current_tokens = sum(token for token, _ in self.token_usage)

    def add_tokens(self, tokens):
        """
        Add tokens to the token usage queue.

        Args:
            tokens (int): Number of tokens to be added.
        """
        self.token_usage.append((tokens, time.time()))


async def fetch(
    session, url, headers, payload, rate_limiter, pbar=None, logit_bias=None
):
    """
    Asynchronous function to fetch data using POST request with retries for
    rate limits.

    Args:
        session (aiohttp.ClientSession): The session to make the request.
        url (str): The URL endpoint to fetch from.
        headers (dict): Headers to include in the request.
        payload (dict): Data payload to send with the request.
        rate_limiter (RateLimiter): An instance of RateLimiter to control the
                                    request rate.
        pbar (optional): A progress bar instance to show progress. Defaults to
                        None.
        logit_bias (optional): Bias for the logit. Defaults to None.

    Returns:
        dict: The response data if the request was successful.
    """
    while True:
        await rate_limiter.wait()  # Wait for the rate limiter

        if logit_bias:
            payload['logit_bias'] = logit_bias

        async with session.post(
            url, headers=headers, data=json.dumps(payload)
        ) as response:
            response_data = await response.json()
            if response.status == 429:  # Rate limit exceeded
                print("Rate limit exceeded, sleeping...")
                await asyncio.sleep(SECONDS_TO_PAUSE_AFTER_RATE_LIMIT_ERROR)
                continue  # Retry the request

            choices = response_data.get('choices')
            total_tokens = response_data.get('usage',
                                             {}).get('total_tokens', 0)
            rate_limiter.add_tokens(
                total_tokens
            )  # Add the tokens used to the rate limiter

            if choices and len(choices) > 0 and choices[0]:
                if pbar:
                    pbar.update(1)
                return response_data

            print(f"Invalid choices in response. Choices: {choices}")

        print("Response criteria not met, retrying...")
        continue


async def parallel_completions(
    message_batches,
    model,
    max_tokens,
    temperature=1.3,
    presence_penalty=0,
    pbar=None,
    logit_bias=None
):
    """
    Asynchronous function to perform parallel completions using OpenAI's API.

    Args:
        message_batches (list): A list containing batches of messages for
                                completion.
        model (str): The model to be used for completion.
        max_tokens (int): Maximum tokens to be used for completion.
        temperature (float, optional): Sampling temperature. Defaults to 1.3.
        presence_penalty (float, optional): Presence penalty for completion.
                                            Defaults to 0.
        pbar (optional): A progress bar instance to show progress. Defaults to
                        None.
        logit_bias (optional): Bias for the logit. Defaults to None.

    Returns:
        list: A list of responses containing completions for each message
            batch.
    """
    url = "https://api.openai.com/v1/chat/completions"
    headers = {
        "Authorization": f"Bearer {openai.api_key}",
        "Content-Type": "application/json"
    }

    rate_limiter = RateLimiter(
        MAX_REQUESTS_PER_MINUTE / 60, MAX_TOKENS_PER_MINUTE
    )  # Create a rate limiter

    proxy = os.environ.get("all_proxy")
    if proxy:
        connector = ProxyConnector.from_url(proxy)
    else:
        connector = None

    async with aiohttp.ClientSession(connector=connector) as session:
        tasks = [
            asyncio.ensure_future(
                fetch(
                    session, url, headers, {
                        "model": model,
                        "messages": messages,
                        "temperature": temperature,
                        "presence_penalty": presence_penalty,
                        "max_tokens": max_tokens
                    }, rate_limiter, pbar, logit_bias
                )
            ) for messages in message_batches
        ]

        responses = await asyncio.gather(*tasks)

    return responses



"""
This module provides an implementation of a combination enhancer using OpenAI's
model to suggest enhancements.
The primary goal of this module is to leverage the capabilities of OpenAI's
language model to analyze the results of an experiment and provide suggestions
on how to enhance the combination of parameters. This module offers a
prompt-based mechanism, where the language model is prompted with structured
information about the experiment and its results, and the model responds with
potential enhancements.
"""
import copy
from concurrent.futures import ThreadPoolExecutor
from typing import List, Optional

import openai
from tqdm import tqdm

from ..experiment.evaluator import Evaluator
from ..experiment.rate_limiter import RateLimiter
from ..experiment.utils import generate_experiment, get_selection_strategy, run_single_input
from ..logger.token_logger import TokenLogger
from ..result_selectors.selection_context import SelectionContext
from ..schemas.combination_enhancer_configs import OpenAIPromptBasedCombinationEnhancerConfig
from ..schemas.common_structures import InputData
from ..schemas.evaluator_config import OpenAIPromptBasedEvaluatorConfig
from ..schemas.experiment_config import (
    CombinationAggregatedMetrics,
    EnhancerOutput,
    Experiment,
    ExperimentConfig,
    ExperimentResult,
    WrapperConfig,
    WrapperVariation,
)
from ..states.experiment_state import ExperimentState
from .base_combination_enhancer import BaseCombinationEnhancer

PROMPT = """
Given the evaluator results, scales, and descriptions, please analyze and
suggest enhancements to the provided combinations. The combinations are
presented in the format {key: value}, where 'value' can represent prompts or
other parameters. You can define the specifics of the value as needed.
After your analysis, the combination will be re-evaluated by the evaluators,
and feedback will be provided. For context, we will supply the prior
combinations if available. Keey the original {key: value} and keep the
original key, only update the value if needed (keep the orignal value if no
update is needed). The value will be of the same types For example:
{"keyA" : "ValueA", "KeyB": 1} if you think keyB should be and not KeyA output
format should be {"keyA" : "ValueA", "KeyB": 2}, make the outptu so that it
can be pared with pyhon directly using eval
"""

COT = """
First, write out in a step by step manner your reasoning to be sure that your
conclusion is correct. Avoid simply stating the correct answer at the outset.
Reasoning:
"""


def find_best_combination(
    experiment: Experiment
) -> CombinationAggregatedMetrics | None:
    """
    Find the best combination metrics from an experiment. If the experiment
    has selection output, use the best combination from the selection output.
    Otheriwise, use the first combination from the experiment.
    """
    if not experiment.combination_aggregated_metrics or len(
        experiment.combination_aggregated_metrics
    ) == 0:
        return None

    if experiment.selection_output:
        combo_key = experiment.selection_output.best_combination
    else:
        combo_key = experiment.combination_aggregated_metrics[0].combo_key

    for metric in experiment.combination_aggregated_metrics:
        if metric.combo_key == combo_key:
            return metric
    return None


def get_evaluator_config(
    config: ExperimentConfig
) -> List[OpenAIPromptBasedEvaluatorConfig]:
    """
    Retrieve OpenAI prompt-based evaluator resulsts from an experiment
    configuration. Thhis imporver only use the openai_prompt_based_evaluator.
    """
    return [
        c for c in config["evaluators"]  # type: ignore
        if c["name"] == "openai_prompt_based_evaluator"
    ]


def find_evaluator_results(
    configs: List[OpenAIPromptBasedEvaluatorConfig],
    metrics: CombinationAggregatedMetrics | None
) -> str:
    """
    Extract evaluator results as a string from metrics based on provided
    evaluator configurations.
    Args:
        configs (List[OpenAIPromptBasedEvaluatorConfig]):
        Evaluator configurations.
        metrics (CombinationAggregatedMetrics or None): Aggregated metrics of
        a combination.
    Returns:
        str: A string representation of evaluator results.
    """
    res: str = ""
    if not metrics:
        return res
    for k, value in metrics.aggregated_metrics.items():
        for evaluator_config in configs:
            key = evaluator_config["name"]  # type: ignore
            key += ": " + evaluator_config[
                "display_name"] if evaluator_config[  # type: ignore
                    "display_name"] else ""  # type: ignore
            if k == key:
                metric_results = [
                    f"{metric.name},{metric.value}" for metric in value
                ]
                metric_results_string = ', '.join(metric_results)
                res += "\n\n"
                res += "Evaluator Results: " + metric_results_string + "\n"
                res += "Scale: " + evaluator_config["scale_description"
                                                    ] + "\n"  # type: ignore
                res += "Description: " + evaluator_config[
                    "description"] + "\n"  # type: ignore
    return res


def construct_prompt(
    evaluator_outputs: str,
    current_combination: str,
    prior_iterations: Optional[List[str]] = None
) -> str:
    """
    Construct a prompt for OpenAI's model based on evaluator outputs and
    combination history.
    """
    prompt = PROMPT + "\n\n" + evaluator_outputs + "\n\n"
    prompt += "Current Combination: " + current_combination + "\n\n"
    if prior_iterations and len(prior_iterations) > 0:
        prompt += "Prior Iterations: \n\n"
        for iteration in prior_iterations:
            prompt += iteration + "\n\n"
    prompt += COT
    return prompt


def extract_dict_from_string(str_input: str) -> str | None:
    """
    Extract the outermost dictionary from a string, handling nested
    dictionaries.
    """
    open_braces = 0
    dict_start = str_input.find('{')
    for i in range(dict_start, len(str_input)):
        if str_input[i] == '{':
            open_braces += 1
        elif str_input[i] == '}':
            open_braces -= 1
        if open_braces == 0:
            return str_input[dict_start:i + 1]
    return None  # Return None if no matching dictionary found


rate_limiter = RateLimiter(10 / 60)


class OpenAIPromptBasedCombinationEnhancer(BaseCombinationEnhancer):
    """
    Combination enhancer that uses OpenAI's model to enhance the combination.
    """
    default_config = OpenAIPromptBasedCombinationEnhancerConfig(
        name="openai_prompt_based_combination_enhancer",
        openai_model_name="gpt-4",
        max_iterations=3,
        stop_conditions={
            "openai_prompt_based_evaluator: clarity": 2.0,
            "openai_prompt_based_evaluator: relevance": 3,
            "openai_prompt_based_evaluator: catchiness": 3
        }
    )

    def __init__(self, config: OpenAIPromptBasedCombinationEnhancerConfig):
        super().__init__(config)
        self.config = config

    def parallel_task(self, data, all_combinations, logger, evaluator):
        """
        Execute a single input run in parallel.
        """
        rate_limiter()
        return run_single_input(
            data,
            self.updated_config,
            all_combinations=all_combinations,
            logger=logger,
            evaluator=evaluator
        )

    def check_if_done(self, experiment: Experiment) -> bool:
        """
        Check if the iterative enhancement process should stop.

        Evaluates predefined stop conditions or average score thresholds to
        determine if the iterative process should halt.
        """

        combo = experiment.combination_aggregated_metrics
        condition_met = []
        average_value: float = 0.0
        for k, value in combo[0].aggregated_metrics.items():
            if self.config.stop_conditions and k in self.config.stop_conditions:  # type: ignore
                average_value += value[
                    0].value  # Assume we only have one metrics calculation.
                if self.config.stop_conditions[k] <= value[  # type: ignore
                    0].value:
                    condition_met.append(True)
        if len(condition_met) > 0 and len(condition_met) == len(
            self.config.stop_conditions  # type: ignore
        ):
            return True
        if self.config.average_score and self.config.average_score <= average_value / len( # type: ignore
            combo[0].aggregated_metrics
        ):
            return True
        return False

    def enhance(
        self, experiment: Experiment, config: ExperimentConfig,
        evaluator: Evaluator, token_logger: TokenLogger
    ) -> EnhancerOutput:
        experiments: List[Experiment] = []
        prior_iterations: List[str] = []
        self.updated_config = copy.deepcopy(config)
        original_combo_key = ''
        results: List[ExperimentResult] = []
        data: List[InputData] = []
        self.updated_config["variations"] = []  # type: ignore
        for i in range(self.config.max_iterations):  # type: ignore
            print(f"Start iteration {i}...")
            current_iteration_results: List[ExperimentResult] = []
            configs = get_evaluator_config(config)
            combo = find_best_combination(experiment)
            if not combo:
                continue
            if len(results) == 0 and combo:
                original_combo_key = combo.combo_key
                for r in combo.experiment_results:
                    r.input_data.content.pop("raw_output", None)
                    results.append(r)
                results.extend(combo.experiment_results)
            evaluator_result = find_evaluator_results(configs, combo)
            message = [{
                "role":
                "user",
                "content":
                construct_prompt(
                    evaluator_result, combo.combo_key, prior_iterations
                )
            }]
            prior_iterations.append(combo.combo_key)
            while True:
                try:

                    response = openai.ChatCompletion.create(
                        model="gpt-4", messages=message, max_tokens=5000
                    )
                    exper = extract_dict_from_string(
                        response["choices"][0]["message"]["content"]
                    )
                    if exper:
                        extracted_dict = eval(exper)
                    else:
                        continue
                    break
                except Exception as exception:
                    print("Retrying: ", exception)

            for k, v in extracted_dict.items():
                # TODO: Support custom class value_type
                self.updated_config["variations"].append(( # type: ignore
                    WrapperConfig(
                        name=k,
                        variations=[
                            WrapperVariation(
                                value=v, value_type=str(type(v)).split("'")[1]
                            )
                        ]
                    )
                ))
            state = ExperimentState.get_instance()
            state.clear_variations_for_experiment()
            state.set_experiment_config(self.updated_config)
            state.active = True
            all_combinations = state.get_all_variation_combinations()
            total = len(combo.experiment_results)
            data.clear()
            for r in combo.experiment_results:
                input_data = copy.deepcopy(r.input_data)
                input_data.content.pop("raw_output", None)
                data.append(input_data)
            with tqdm(total=total, desc="Processing", unit="item") as pbar:
                with ThreadPoolExecutor() as executor:
                    for res in executor.map(
                        self.parallel_task, data,
                        [all_combinations] * len(data),
                        [token_logger] * len(data), [evaluator] * len(data)
                    ):
                        current_iteration_results.extend(res)
                        pbar.update(len(res))
            experiment = generate_experiment(
                current_iteration_results,
                evaluator,
                evaluate_group=False,
                evaluate_all=False
            )
            experiments.append(experiment)
            if self.check_if_done(experiment):
                break

        for exp in experiments:
            for res in exp.combination_aggregated_metrics:
                tmp = []
                for r in res.experiment_results:
                    r.input_data.content.pop("raw_output", None)
                    tmp.append(r)
                results.extend(res.experiment_results)
        experiment = generate_experiment(
            results, evaluator, evaluate_group=False, evaluate_all=False
        )

        enable_custom_func = "custom_function" in self.config  #type: ignore
        strategy = get_selection_strategy(self.updated_config)
        if strategy and enable_custom_func:
            context_trade_off = SelectionContext(strategy=strategy)
            experiment.selection_output = context_trade_off.execute_selection( # type: ignore
                experiment=experiment
            )
        enhancer_output = EnhancerOutput(
            group_experiment_results=experiment.group_experiment_results,
            combination_aggregated_metrics=experiment.
            combination_aggregated_metrics,
            original_best_combo_key=original_combo_key,
            selection_output=experiment.selection_output
        )
        return enhancer_output


BaseCombinationEnhancer.register_enhancer(
    "openai_prompt_based_combination_enhancer",
    OpenAIPromptBasedCombinationEnhancer,
    OpenAIPromptBasedCombinationEnhancerConfig
)



import json
import os
import pickle
import time
from collections import defaultdict
from typing import Dict, List

import numpy as np
import openai
import tiktoken

from yival.dataset.data_utils import (
    evaluate_condition,
    read_code_from_path_or_module,
    transform_experiment_result_generic,
)
from yival.schemas.experiment_config import Experiment, ExperimentResult

encoding = tiktoken.get_encoding("cl100k_base")

TARGET_EPOCHS = 3
MIN_TARGET_EXAMPLES = 100
MAX_TARGET_EXAMPLES = 25000
MIN_DEFAULT_EPOCHS = 1
MAX_DEFAULT_EPOCHS = 25


def num_tokens_from_messages(
    messages, tokens_per_message=3, tokens_per_name=1
):
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3
    return num_tokens


def num_assistant_tokens_from_messages(messages):
    num_tokens = 0
    for message in messages:
        if message["role"] == "assistant":
            num_tokens += len(encoding.encode(message["content"]))
    return num_tokens


def _print_distribution(values, name):
    print(f"\n#### Distribution of {name}:")
    print(f"min / max: {min(values)}, {max(values)}")
    print(f"mean / median: {np.mean(values)}, {np.median(values)}")
    print(f"p5 / p95: {np.quantile(values, 0.1)}, {np.quantile(values, 0.9)}")


def _print_stats(dataset: list[dict[str, list[dict[str, str]]]]):
    n_missing_system = 0
    n_missing_user = 0
    n_messages = []
    convo_lens = []
    assistant_message_lens = []
    MAX_TOKENS_PER_EXAMPLE = 4096

    for ex in dataset:
        messages = ex["messages"]
        if not any(message["role"] == "system" for message in messages):
            n_missing_system += 1
        if not any(message["role"] == "user" for message in messages):
            n_missing_user += 1
        n_messages.append(len(messages))
        convo_lens.append(num_tokens_from_messages(messages))
        assistant_message_lens.append(
            num_assistant_tokens_from_messages(messages)
        )

    print("Num examples missing system message:", n_missing_system)
    print("Num examples missing user message:", n_missing_user)
    _print_distribution(n_messages, "num_messages_per_example")
    _print_distribution(convo_lens, "num_total_tokens_per_example")
    _print_distribution(
        assistant_message_lens, "num_assistant_tokens_per_example"
    )
    n_too_long = sum(l > 4096 for l in convo_lens)
    print(
        f"\n{n_too_long} examples may be over the 4096 token limit, they will be truncated during fine-tuning"
    )

    n_epochs = TARGET_EPOCHS
    n_train_examples = len(dataset)
    if n_train_examples * TARGET_EPOCHS < MIN_TARGET_EXAMPLES:
        n_epochs = min(
            MAX_DEFAULT_EPOCHS, MIN_TARGET_EXAMPLES // n_train_examples
        )
    elif n_train_examples * TARGET_EPOCHS > MAX_TARGET_EXAMPLES:
        n_epochs = max(
            MIN_DEFAULT_EPOCHS, MAX_TARGET_EXAMPLES // n_train_examples
        )

    n_billing_tokens_in_dataset = sum(
        min(MAX_TOKENS_PER_EXAMPLE, length) for length in convo_lens
    )
    print(
        f"Dataset has ~{n_billing_tokens_in_dataset} tokens that will be charged for during training"
    )
    print(f"By default, you'll train for {n_epochs} epochs on this dataset")
    print(
        f"By default, you'll be charged for ~{n_epochs * n_billing_tokens_in_dataset} tokens"
    )


def validate_message(dataset) -> bool:
    # Format error checks
    format_errors: dict[str, int] = defaultdict(int)

    for ex in dataset:
        if not isinstance(ex, dict):
            format_errors["data_type"] += 1
            continue

        messages = ex.get("messages", None)
        if not messages:
            format_errors["missing_messages_list"] += 1
            continue

        for message in messages:
            if "role" not in message or "content" not in message:
                format_errors["message_missing_key"] += 1

            if any(k not in ("role", "content", "name") for k in message):
                format_errors["message_unrecognized_key"] += 1

            if message.get("role",
                           None) not in ("system", "user", "assistant"):
                format_errors["unrecognized_role"] += 1

            content = message.get("content", None)
            if not content or not isinstance(content, str):
                format_errors["missing_content"] += 1

        if not any(
            message.get("role", None) == "assistant" for message in messages
        ):
            format_errors["example_missing_assistant_message"] += 1

    if format_errors:
        print("Found errors:")
        for k, v in format_errors.items():
            print(f"{k}: {v}")
        return False
    else:
        print("No errors found")
        return True


def _format_data_for_chatgpt_finetune(
    data: List[Dict[str, str]],
    system_prompt: str = "",
) -> List[Dict[str, List[Dict[str, str]]]]:
    """
    Transforms the input data to the desired data for finetune.

    Args:
    - system_prompt (str): System message to include at the start of each chat.
    - data (List[Dict[str, str]]): List of data points with 'Input' and 'Output' keys.

    Returns:
    - List[Dict[str, List[Dict[str, str]]]]: Formatted data.
    """
    formatted_data = []

    for entry in data:
        chat_entry: dict[str, list[dict[str, str]]] = {"messages": []}
        if system_prompt:
            chat_entry["messages"].append({
                "role": "system",
                "content": system_prompt
            })
        elif "Instruction" in entry:
            chat_entry["messages"].append({
                "role": "system",
                "content": entry["Instruction"]
            })
        chat_entry["messages"].append({
            "role": "user",
            "content": entry["Input"]
        })
        chat_entry["messages"].append({
            "role": "assistant",
            "content": entry["Output"]
        })
        formatted_data.append(chat_entry)

    return formatted_data


def finetune(
    input_file: str,
    condition: str,
    custom_function: str,
    system_prompt: str = "",
    model_suffx: str = "",
) -> str:
    """
    Fine-tunes a gpt-3.5 using provided data and conditions.

    Args:
    - input_file (str): Path to the input file containing experiment results.
    - condition (str): Condition to evaluate for extracting relevant results.
    - custom_function (str): Path or module containing the custom function used in the experiment.
    - system_prompt (str, optional): System message to prepend to each chat. Defaults to None.
    - model_suffix: (str, optional): Suffix to append to the model name. Defaults to None.

    Returns:
    - str: ID of the fine-tuned model.
    """

    code = read_code_from_path_or_module(custom_function)
    if not code:
        print("Failed to read code from path or module.")

        return ""

        return
    with open(input_file, 'rb') as f:
        result: Experiment = pickle.load(f)
    result_pairs = []
    for combo_result in result.combination_aggregated_metrics:
        results: List[ExperimentResult] = combo_result.experiment_results
        for rs in results:
            if rs.evaluator_outputs:
                for evaluator_output in rs.evaluator_outputs:
                    condition_met = evaluate_condition(
                        condition, evaluator_output
                    )
                    if condition_met:
                        result_pair = transform_experiment_result_generic(
                            code, rs
                        )
                        result_pairs.append(result_pair)

    formatted_data = _format_data_for_chatgpt_finetune(
        result_pairs, system_prompt
    )
    if validate_message(formatted_data):
        openai.api_key = os.getenv("OPENAI_API_KEY")
        _print_stats(formatted_data)
        with open("tmp.jsonl", "w") as f:
            for entry in formatted_data:
                json_str = json.dumps(entry)
                f.write(json_str + "\n")

        with open("tmp.jsonl", "rb") as f:
            output = openai.File.create(
                file=f,
                purpose="fine-tune",
            )
            print(output)
            while True:
                try:
                    model_details = openai.FineTuningJob.create(
                        training_file=output["id"],
                        model="gpt-3.5-turbo",
                        suffix=model_suffx if model_suffx else ""
                    )
                    print("Fine-tuning job started: %s", model_details["id"])
                    break
                except openai.error.InvalidRequestError as e:
                    print(e)
                    print("Waiting for file to be ready...")
                    time.sleep(60)

            ft_job = openai.FineTuningJob.retrieve(model_details["id"])
            print(ft_job)
            while ft_job["status"] != "succeeded" and ft_job["status"
                                                             ] != "failed":
                ft_job = openai.FineTuningJob.retrieve(model_details["id"])

                print(ft_job)
                time.sleep(60)

            print("Fine-tuning job finished: %s", model_details["id"])
            return ft_job["fine_tuned_model"]
        return ""
    else:
        print("Something is wrong, please check the data format.")
        return ""


def main():
    finetune(
        'test_demo_results.pkl',
        "name == openai_prompt_based_evaluator AND result >= 0 AND display_name == clarity",
    )


if __name__ == "__main__":
    main()



class TokenLogger:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.current_usage = 0
        return cls._instance

    def log(self, tokens: int):
        self.current_usage += tokens

    def get_current_usage(self) -> int:
        """Get the token usage of the current operation."""
        return self.current_usage

    def reset(self):
        self.current_usage = 0



from enum import Enum
from typing import Any, Dict, List, Optional, Type

import yaml

from ..data.base_reader import BaseReader
from ..data_generators.base_data_generator import BaseDataGenerator
from ..enhancers.base_combination_enhancer import BaseCombinationEnhancer
from ..evaluators.base_evaluator import BaseEvaluator
from ..result_selectors.selection_strategy import SelectionStrategy
from ..schemas.experiment_config import WrapperConfig
from ..variation_generators.base_variation_generator import BaseVariationGenerator
from ..wrappers.base_wrapper import BaseWrapper


def recursive_asdict(obj) -> Any:
    if isinstance(obj, list):
        return [recursive_asdict(item) for item in obj]
    elif isinstance(obj, dict):
        return {key: recursive_asdict(value) for key, value in obj.items()}
    elif isinstance(obj, Enum):  # Handle Enum values
        return obj.value
    elif hasattr(obj, 'asdict'):  # For dataclasses with asdict
        return obj.asdict()
    else:
        return obj


def get_default_config(component_class: Type[Any]) -> Optional[Dict[str, Any]]:
    """Retrieve the default configuration of a component, if available."""
    default = getattr(component_class, "default_config", None)
    if not isinstance(default, dict):
        return default.__dict__
    return default


def get_config_for_component(component_name, get_component_func):
    """Get configuration for a given component."""
    component_cls = get_component_func(component_name)
    if component_cls:
        default_config = get_default_config(component_cls)
        if default_config:
            return {component_name: default_config}
        else:
            return {component_name: {}}
    return {}


def generate_dataset_section(source_type, reader_name,
                             data_generator_names) -> Dict[str, Any]:
    """Generate the dataset section of the experiment config."""
    dataset_section = {"source_type": source_type}
    if source_type == "dataset":
        dataset_section["file_path"] = "/path/to/file_path"
    if reader_name and source_type == "dataset":
        dataset_section["reader"] = reader_name
        reader_config = get_config_for_component(
            reader_name, BaseReader.get_reader
        )
        if reader_config:
            dataset_section["reader_config"] = reader_config
    if data_generator_names and source_type == "machine_generated":
        data_generators_section = {}
        for name in data_generator_names:
            data_generator_config = get_config_for_component(
                name, BaseDataGenerator.get_data_generator
            )
            if data_generator_config:
                data_generators_section.update(data_generator_config)
        if data_generators_section:
            dataset_section["data_generators"] = data_generators_section
    return dataset_section


def generate_enhancer_config(enhancer_name) -> Optional[Dict[str, Any]]:
    """Generate the enhanced section of the experiment config."""
    return get_config_for_component(
        enhancer_name, BaseCombinationEnhancer.get_enhancer
    )


def generate_variations_section(wrapper_configs) -> str:
    """Generate the variations section."""
    if not wrapper_configs:
        return """
        # variations:
        #   - name: wrapper_name
        #     variations:
        #       - value_type: str
        #         value: "example_variation"
        #         instantiated_value: "example_variation"
        #     generator_name: example_generator_name
        #     generator_config:
        #       number_of_variations: 5
        #       output_path: /path/to/output
        """

    variations_list = []
    for wrapper_config in wrapper_configs:
        wrapper_dict = wrapper_config.asdict()
        if wrapper_config.generator_name:
            wrapper_generator_cls = BaseVariationGenerator.get_variation_generator(
                wrapper_config.generator_name
            )
            if wrapper_generator_cls:
                default_config = get_default_config(wrapper_generator_cls)
                if default_config:
                    wrapper_dict["generator_config"] = default_config
        variations_list.append(wrapper_dict)
    res = "\nvariations:\n"
    res += yaml.safe_dump(
        variations_list,
        default_flow_style=False,
        allow_unicode=True,
        indent=2
    )
    return res


# Refactored main function
def generate_experiment_config_yaml(
    custom_function: str,
    source_type: str = "dataset",
    evaluator_names: Optional[List[str]] = None,
    reader_name: Optional[str] = None,
    enhancer_name: Optional[str] = None,
    wrapper_names: Optional[List[str]] = None,
    data_generator_names: Optional[List[str]] = None,
    selection_strategy_name: Optional[str] = None,
    wrapper_configs: Optional[List[WrapperConfig]] = None,
    custom_reader: Optional[Dict[str, Dict[str, Any]]] = None,
    custom_wrappers: Optional[Dict[str, Dict[str, Any]]] = None,
    custom_evaluators: Optional[Dict[str, Dict[str, Any]]] = None,
    custom_data_generators: Optional[Dict[str, Dict[str, Any]]] = None,
    custom_variation_generators: Optional[Dict[str, Dict[str, Any]]] = None,
    custom_selection_strategy: Optional[Dict[str, Dict[str, Any]]] = None,
    custom_enhancer: Optional[Dict[str, Dict[str, Any]]] = None
) -> str:

    experiment_config = {
        "description":
        "Generated experiment config",
        "dataset":
        generate_dataset_section(
            source_type, reader_name, data_generator_names
        ),
        "custom_function":
        custom_function
    }

    if enhancer_name:
        config = generate_enhancer_config(enhancer_name)
        if config:
            experiment_config["enhancer"] = config

    # Add selection strategy
    if selection_strategy_name:
        experiment_config["selection_strategy"] = get_config_for_component(
            selection_strategy_name, SelectionStrategy.get_strategy
        )

    # Add custom sections if they exist
    custom_sections = {
        "custom_reader": custom_reader,
        "custom_enhancer": custom_enhancer,
        "custom_wrappers": custom_wrappers,
        "custom_evaluators": custom_evaluators,
        "custom_data_generators": custom_data_generators,
        "custom_variation_generators": custom_variation_generators,
        "custom_selection_strategies": custom_selection_strategy
    }
    for section_name, section_content in custom_sections.items():
        if section_content:
            experiment_config[section_name] = section_content

    # Process evaluators
    if evaluator_names:
        evaluators_section = []
        for name in evaluator_names:
            evaluator_config = get_config_for_component(
                name, BaseEvaluator.get_evaluator
            )
            if evaluator_config:
                evaluator_config["name"] = name  # Add the name field
                evaluators_section.append(evaluator_config)
        experiment_config["evaluators"] = evaluators_section

    # Process wrappers
    if wrapper_names:
        wrappers_section = {}
        for name in wrapper_names:
            wrapper_config = get_config_for_component(
                name, BaseWrapper.get_wrapper
            )
            wrappers_section.update(wrapper_config)
        experiment_config["wrapper_configs"] = wrappers_section

    experiment_config = recursive_asdict(experiment_config)
    yaml_string = yaml.safe_dump(
        experiment_config, default_flow_style=False, allow_unicode=True
    )

    if not evaluator_names:
        yaml_string += "\n# evaluators: []\n"
    if not wrapper_names:
        yaml_string += "\n# wrapper_configs: {}\n"

    yaml_string += generate_variations_section(wrapper_configs)
    yaml_string = (
        "# This is a generated template. Modify the values as needed.\n\n" +
        yaml_string
    )
    return yaml_string



"""
Demo code for question answering using GPT-3.
"""
import os

import openai

from yival.logger.token_logger import TokenLogger
from yival.schemas.experiment_config import MultimodalOutput
from yival.states.experiment_state import ExperimentState
from yival.wrappers.string_wrapper import StringWrapper


def qa(question: str, state: ExperimentState) -> MultimodalOutput:
    """
    Demo code for question answering using GPT-3.
    """
    logger = TokenLogger()
    logger.reset()
    # Ensure you have your OpenAI API key set up
    openai.api_key = os.getenv("OPENAI_API_KEY")

    # Create a chat message sequence
    prompt = str(StringWrapper("", name="qa", state=state))
    messages = [{
        "role":
        "system",
        "content":
        "You are a helpful assistant that will answer the question with only option."
    }, {
        "role": "user",
        "content": f'{question} ' + prompt
    }]
    # Use the chat-based completion
    response = openai.ChatCompletion.create(
        model="gpt-3.5-turbo", messages=messages, temperature=0
    )

    # Extract the assistant's message (translated text) from the response
    answer = MultimodalOutput(
        text_output=response['choices'][0]['message']['content'],
    )
    token_usage = response['usage']['total_tokens']
    logger.log(token_usage)

    return answer



# type: ignore

import ast
import base64
import hashlib
import io
import json
import os
import re
import textwrap
import urllib.parse
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, List, Optional

import dash  # type: ignore
import dash_bootstrap_components as dbc  # type: ignore
import numpy as np
import pandas as pd  # type: ignore
import plotly.express as px  # type: ignore
from dash import dash_table, dcc, html  # type: ignore
from dash.dependencies import ALL, MATCH, Input, Output, State
from dash_dangerously_set_inner_html import DangerouslySetInnerHTML
from PIL import Image
from pyngrok import ngrok

from yival.experiment.rate_limiter import RateLimiter
from yival.experiment.utils import (
    call_function_from_string,
    generate_experiment,
    get_function_args,
    run_single_input,
)
from yival.schemas.experiment_config import (
    CombinationAggregatedMetrics,
    EvaluatorOutput,
    Experiment,
    ExperimentConfig,
    ExperimentResult,
    GroupedExperimentResult,
    MultimodalOutput,
)

from ...schemas.common_structures import InputData
from .hexagram import HEXAGRAMS, generate_hexagram_figure
from .utils import (
    generate_group_key_combination_data,
    generate_heatmap_style,
    generate_legend,
    sanitize_column_name,
    sanitize_group_key,
)


def include_image_base64(data_dict):
    """Check if a string includes a base64 encoded image."""
    pattern = r'iVBORw.+'
    for item in data_dict:
        for record in item:
            for value in record.values():
                if re.search(pattern, value):
                    return True
    return False


def include_video(data_dict):
    """Check if a string includes a video."""
    pattern = r'<yival_video_output>'
    for item in data_dict:
        for record in item:
            for value in record.values():
                if re.search(pattern, value):
                    return True
    return False


def is_base64_image(value):
    """Check if a string is a base64 encoded image."""
    pattern = r'iVBORw.+'
    return bool(re.match(pattern, value))


def base64_to_img(base64_string):
    """Convert a base64 string into a PIL Image."""
    decoded = base64.b64decode(base64_string)
    image = Image.open(io.BytesIO(decoded))
    return image


def extract_and_decode_image_from_string(data_string):
    """Extract and decode the first image from a string include base64 encoded and return a dictionary."""
    # Extract text_output, image_output and evaluate part
    image_output_string_match = re.search(
        r"\['(.*?)',", data_string, re.DOTALL
    )
    if image_output_string_match:
        text_output_match = re.search(
            r'<yival_raw_output>\n(.*?)\n</yival_raw_output>', data_string,
            re.DOTALL
        )
        text_output = text_output_match.group(1)
        image_output_string = image_output_string_match.group(1)
        evaluate_match = re.search(r"\](.*)", data_string, re.DOTALL)
        evaluate = evaluate_match.group(1).strip()
        # Check if image_output_string is base64 image
        image_output = base64_to_img(image_output_string)
        return {
            "text_output": text_output,
            "image_output": image_output,
            "evaluate": evaluate
        }
    else:
        return None


def extract_and_decode_video_from_string(data_string):
    """Extract and decode the first video from a string include video urls and return a dictionary."""
    # Extract text_output, video_output and evaluate part
    video_output_string_match = re.search(
        r"<yival_video_output>(.*?)</yival_video_output>", data_string,
        re.DOTALL
    )
    if video_output_string_match:
        text_output_match = re.search(
            r'<yival_raw_output>\n(.*?)\n</yival_raw_output>', data_string,
            re.DOTALL
        )
        text_output = text_output_match.group(1)
        video_output = video_output_string_match.group(1)
        evaluate_match = re.search(
            r"</yival_video_output>(.*)", data_string, re.DOTALL
        )
        evaluate = evaluate_match.group(1).strip()
        return {
            "text_output": text_output,
            "video_output": video_output,
            "evaluate": evaluate
        }
    else:
        return None


def extract_and_decode_image(data_dict):
    """Extract and decode image from a dictionary include base64 encoded."""
    new_data_dict = []
    for item in data_dict:
        for record in item:
            new_record = {}
            for key, value in record.items():
                if isinstance(value, str):
                    image = extract_and_decode_image_from_string(value)
                    if image is not None:
                        new_record[key] = image
                    else:
                        new_record[key] = value
                else:
                    new_record[key] = value
            new_data_dict.append(new_record)
    return new_data_dict


def extract_and_decode_video(data_dict):
    """Extract and decode video from a dictionary include video url."""
    new_data_dict = []
    for item in data_dict:
        for record in item:
            new_record = {}
            for key, value in record.items():
                if isinstance(value, str):
                    video = extract_and_decode_video_from_string(value)
                    if video is not None:
                        new_record[key] = video
                    else:
                        new_record[key] = value
                else:
                    new_record[key] = value
            new_data_dict.append(new_record)
    return new_data_dict


def create_table(data):
    """Create an HTML table from a list of dictionaries, where the values can be strings or PIL images."""
    # Create the header row
    header_row = [
        html.Th(key if key != "Hashed Group Key" else "Human Rating")
        for key in data[0].keys()
    ]
    table_rows = [html.Tr(header_row)]

    # Create the data rows
    for record in data:
        row = []
        for key, value in record.items():
            if key == "Hashed Group Key":
                # Convert the hashed group key to a hyperlink
                href = f"/rating-result/{value}"
                cell = html.Td(
                    dcc.Link("Link", href=href)
                )  # Replace "Link text" with the text you want to display
            elif isinstance(value, dict):
                img = html.Img(
                    src=pil_image_to_base64(value["image_output"]),
                    style={
                        'width': '200px',
                        'height': '200px'
                    }
                )
                text = html.P(value["text_output"])
                cell = html.Td([
                    text, html.Br(), img,
                    html.Br(), value["evaluate"]
                ])
            else:
                cell = html.Td(value)
            row.append(cell)
        table_rows.append(html.Tr(row))
    return table_rows


def create_video_table(data):
    """Create an HTML table from a list of dictionaries, where the values can be strings or video."""
    # Create the header row
    header_row = [
        html.Th(key if key != "Hashed Group Key" else "Human Rating")
        for key in data[0].keys()
    ]
    table_rows = [html.Tr(header_row)]

    # Create the data rows
    for record in data:
        row = []
        for key, value in record.items():
            if key == "Hashed Group Key":
                # Convert the hashed group key to a hyperlink
                href = f"/rating-result/{value}"
                cell = html.Td(
                    dcc.Link("Link", href=href)
                )  # Replace "Link text" with the text you want to display on UI
            elif isinstance(value, dict):
                text = html.P(value["text_output"])
                video = html.Div(
                    children=html.Video(
                        controls=True,
                        id='movie_player',
                        src=value["video_output"],
                        autoPlay=False
                    ),
                    style={
                        'height': '200px',
                        'width': '200px',
                    }
                )
                cell = html.Td([
                    text, html.Br(), video,
                    html.Br(), value["evaluate"]
                ])
            else:
                cell = html.Td(value)
            row.append(cell)
        table_rows.append(html.Tr(row))
    return table_rows


def pil_image_to_base64(image: Image, format: str = "PNG") -> str:
    buffered = io.BytesIO()
    image.save(buffered, format=format)
    img_str = base64.b64encode(buffered.getvalue())
    return "data:image/png;base64," + img_str.decode()


def handle_output(output):
    if isinstance(output, list):
        if all(isinstance(item, Image.Image) for item in output):
            return [
                html.Img(src=pil_image_to_base64(img), className="image")
                for img in output
            ]
        else:
            return [html.P(str(item)) for item in output]
    else:
        return html.P(str(output))


def df_to_table(df):
    return html.Table([html.Tr([html.Th(col) for col in df.columns])] + [
        html.Tr([html.Td(row[col]) for col in df.columns])
        for index, row in df.iterrows()
    ])


def create_dash_app(
    experiment_data: Experiment, experiment_config: ExperimentConfig,
    function_args: Dict[str, Any], all_combinations, state, logger, evaluator,
    interactive_mode
):

    def parallel_task(data_point, all_combinations, logger, evaluator):
        """Task to be run in parallel for processing data points."""
        RateLimiter(60 / 60)()  # Ensure rate limit
        return run_single_input(
            data_point,
            experiment_config,
            all_combinations=all_combinations,
            logger=logger,
            evaluator=evaluator
        )

    def get_group_experiment_result_from_hash(hashed_group_key):
        for group_result in experiment_data.group_experiment_results:
            group_key = sanitize_group_key(group_result.group_key)
            if hashlib.sha256(group_key.encode()
                              ).hexdigest() == hashed_group_key:
                return group_result
        for group_result in experiment_data.enhancer_output.group_experiment_results:
            group_key = sanitize_group_key(group_result.group_key)
            if hashlib.sha256(group_key.encode()
                              ).hexdigest() == hashed_group_key:
                return group_result
        return None

    def generate_navigation():
        return dbc.NavbarSimple(
            children=[
                dbc.NavItem(
                    dbc.NavLink(
                        "Experiment Results Analysis",
                        href="/experiment-results"
                    )
                ),
                dbc.NavItem(
                    dbc.NavLink(
                        "Detailed Experiment Results", href="/group-key-combo"
                    )
                ),
                dbc.NavItem(
                    dbc.NavLink(
                        "Enhancer Experiment Results Analysis",
                        href="/enhancer-experiment-results"
                    )
                ),
                dbc.NavItem(
                    dbc.NavLink(
                        "Enhancer Detailed Experiment Results",
                        href="/enhancer-group-key-combo"
                    )
                ),
                dbc.NavItem(
                    dbc.Button(
                        "Export Data",
                        id="export-btn",
                        color="primary",
                        className="ml-2"
                    )
                ),
                dbc.NavItem(dbc.NavLink(
                    "Playground",
                    href="/interactive",
                )),
                dbc.NavItem(
                    dbc.NavLink("Data Analysis", href="/data-analysis")
                ),
                dbc.NavItem(
                    dbc.NavLink(
                        "Use Best Combinations",
                        href="/use-best-result",
                    )
                ),
            ],
            brand="YiVal",
            brand_href="/",
            color="primary",
            dark=True,
        )

    def index_page():
        return html.Div([
            html.H1("Yijing (I Ching)"),
            html.Button('Cast Your Fortune', id='cast-fortune-btn'),
            html.Div(id='hexagram-container')
        ])

    def generate_combo_metrics_data(
        combo_metrics: List[CombinationAggregatedMetrics],
        group_experiment_results: List[GroupedExperimentResult]
    ) -> pd.DataFrame:
        data = []
        for metric in combo_metrics:
            row = {
                "Hyperparameters":
                "\n".join(
                    textwrap.wrap(
                        str(metric.combo_key).replace('"',
                                                      "").replace("'", ""), 90
                    )
                )
            }

            for k, v in metric.aggregated_metrics.items():
                row[k] = ', '.join([f"{m.name}: {m.value}" for m in v])
            row['Average Token Usage'] = str(metric.average_token_usage)
            row['Average Latency'] = str(metric.average_latency)

            if metric.combine_evaluator_outputs:
                for e in metric.combine_evaluator_outputs:
                    column_name = f"{e.name} Output"
                    if e.display_name:
                        column_name += f" ({e.display_name})"
                    row[f"{e.name} Output"] = e.result

            sample_count = 0
            for group in group_experiment_results:
                matching_results = [
                    exp_result.raw_output
                    for exp_result in group.experiment_results
                    if json.dumps(exp_result.combination) == metric.combo_key
                ]
                if matching_results:

                    json_str = group.group_key.replace(
                        'example_id:', '"example_id":'
                    ).replace('content:', '"content":').replace(
                        'expected_result:', '"expected_result":'
                    )
                    escaped_json_str = json_str.replace("\n", "\\n"
                                                        ).replace("\t", "\\t")
                    valid_json_str = escaped_json_str.replace(
                        ": None", ": null"
                    )
                    try:
                        data_dict = json.loads(valid_json_str)
                        content = data_dict['content']
                        if content:
                            items = [str(value) for value in content.values()]
                            group_key = ", ".join([
                                item.strip() for item in items
                            ])
                        else:
                            group_key = ""
                    except Exception as e:
                        group_key = group.group_key
                    group_key = sanitize_column_name(group_key)
                    # if sample_count < 3:
                    #     row[f"Sample {sample_count + 1} ({group_key})"
                    #         ] = matching_results[0]
                    #     sample_count += 1
                    # else:
                    #     break
                    row[f"Sample {sample_count + 1} ({group_key})"
                        ] = matching_results[0]
                    sample_count += 1

            data.append(row)
        for index, row in enumerate(data):
            row["Iteration"] = index
        df = pd.DataFrame(data)
        column_order = ["Iteration"
                        ] + [col for col in df if col != "Iteration"]
        df = df[column_order]
        if 'Average Token Usage' in df:
            df['Average Token Usage'] = pd.to_numeric(
                df['Average Token Usage'], errors='coerce'
            )
        if 'Average Latency' in df:
            df['Average Latency'] = pd.to_numeric(
                df['Average Latency'], errors='coerce'
            )
        return df

    def experiment_results_layout():

        df = generate_combo_metrics_data(
            experiment_data.combination_aggregated_metrics,
            experiment_data.group_experiment_results
        )
        csv_string = df.to_csv(index=False, encoding='utf-8')
        csv_data_url = 'data:text/csv;charset=utf-8,' + urllib.parse.quote(
            csv_string
        )
        sample_columns = [col for col in df.columns if "Sample" in col]
        contains_images = any(
            df[col].apply(
                lambda x: isinstance(x, MultimodalOutput) and x.image_output is
                not None
            ).any() for col in sample_columns
        )
        contains_videos = any(
            df[col].apply(
                lambda x: isinstance(x, MultimodalOutput) and x.video_output is
                not None
            ).any() for col in sample_columns
        )

        multi_div = None

        if contains_images:
            multi_div = image_combo_aggregated_metrics_layout(df)
        elif contains_videos:
            multi_div = video_combo_aggregated_metrics_layout(df)
        else:
            multi_div = combo_aggregated_metrics_layout(df)

        return html.Div([
            html.H3(
                "Experiment Results Analysis", style={'textAlign': 'center'}
            ),
            html.Hr(),
            html.Div([multi_div],
                     style={
                         'overflowY': 'auto',
                         'overflowX': 'auto'
                     }),
            html.Hr(),
            generate_legend(),
            html.A(
                'Export to CSV',
                id='export-link-experiment-results',
                download="experiment_results.csv",
                href=csv_data_url,
                target="_blank"
            ),
            html.Br(),
            dcc.Link('Go to Data Analysis', href='/data-analysis'),
            html.Br(),
            dcc.Link(
                'Go to Detailed Experiment Results', href='/group-key-combo'
            ),
            html.Br(),
            dcc.Link(
                'Go to Enhancer Experiment Results Analysis',
                href='/enhancer-experiment-results'
            ),
            html.Br(),
            dcc.Link(
                'Go to Enhancer Detailed Experiment Results',
                href='/enhancer-group-key-combo'
            ),
            html.Br()
        ])

    def data_analysis_layout():
        return html.Div([
            html.H3("Data Analysis", style={'textAlign': 'center'}),
            analysis_layout(df),
            html.Hr(),
            dcc.Link(
                'Go back to Experiment Results Analysis',
                href='/experiment-results'
            ),
            html.Br(),
            dcc.Link(
                'Go to Detailed Experiment Results', href='/group-key-combo'
            ),
            html.Br(),
            dcc.Link(
                'Go to Enhancer Experiment Results Analysis',
                href='/enhancer-experiment-results'
            ),
            html.Br(),
            dcc.Link(
                'Go to Enhancer Detailed Experiment Results',
                href='/enhancer-group-key-combo'
            ),
            html.Br()
        ])

    def combo_page_layout():
        return html.Div([
            html.H3(
                "Detailed Experiment Results", style={'textAlign': 'center'}
            ),
            html.Hr(),
            group_key_combination_layout(
                experiment_data.group_experiment_results
            ),
            dcc.Link(
                'Go to Enhancer Experiment Results Analysis',
                href='/enhancer-experiment-results'
            ),
            html.Br(),
            dcc.Link(
                'Go to Enhancer Detailed Experiment Results',
                href='/enhancer-group-key-combo'
            ),
            html.Br(),
            html.Hr(),
            html.Div(
                id='current-page-context',
                style={'display': 'none'},
                children='default'
            )
        ])

    def enhancer_experiment_results_layout():
        if not experiment_data.enhancer_output:
            return html.Div([html.H3("No Enhancer Output data available.")])

        df_enhancer = generate_combo_metrics_data(
            experiment_data.enhancer_output.combination_aggregated_metrics,
            experiment_data.enhancer_output.group_experiment_results
        )

        csv_string = df_enhancer.to_csv(index=False, encoding='utf-8')
        csv_data_url = 'data:text/csv;charset=utf-8,' + urllib.parse.quote(
            csv_string
        )

        return html.Div([
            html.H3(
                "Enhancer Experiment Results Analysis",
                style={'textAlign': 'center'}
            ),
            combo_aggregated_metrics_layout(df_enhancer),
            html.Hr(),
            generate_legend(),
            html.A(
                'Export to CSV',
                id='export-link-enhancer-experiment-results',
                download="enhancer_experiment_results.csv",
                href=csv_data_url,
                target="_blank"
            ),
            html.Br(),
            dcc.Link('Go to Data Analysis', href='/data-analysis'),
            html.Br(),
            dcc.Link(
                'Go to Detailed Experiment Results', href='/group-key-combo'
            ),
            html.Br(),
            dcc.Link(
                'Go to Enhancer Detailed Experiment Results',
                href='/enhancer-group-key-combo'
            ),
            html.Br()
        ])

    def analysis_layout(df):
        evaluator_outputs = [
            col for col in df.columns
            if (col != 'Hyperparameters' and 'Sample' not in col)
        ]
        return html.Div([
            html.Div([
                dcc.Dropdown(
                    id='evaluator-dropdown-token',
                    options=[{
                        'label': evaluator,
                        'value': evaluator
                    } for evaluator in evaluator_outputs],
                    value=evaluator_outputs[0] if evaluator_outputs else None,
                    multi=False
                ),
                dcc.Graph(id='comparative-scatter-plot-token')
            ],
                     className="six columns"),
            html.Div([
                dcc.Dropdown(
                    id='evaluator-dropdown-latency',
                    options=[{
                        'label': evaluator,
                        'value': evaluator
                    } for evaluator in evaluator_outputs],
                    value=evaluator_outputs[0] if evaluator_outputs else None,
                    multi=False
                ),
                dcc.Graph(id='comparative-scatter-plot-latency')
            ],
                     className="six columns"),
            html.Div([html.P(id='correlation-coefficient')],
                     className="twelve columns"),
        ],
                        className="row")

    def enhancer_combo_page_layout():
        if not experiment_data.enhancer_output:
            return html.Div([html.H3("No Enhancer Output data available.")])
        return html.Div([
            html.H3(
                "Enhancer Detailed Experiment Results",
                style={'textAlign': 'center'}
            ),
            group_key_combination_layout(
                experiment_data.enhancer_output.group_experiment_results,
                highlight_key=experiment_data.enhancer_output.
                original_best_combo_key
            ),
            dcc.Link(
                'Go to Enhancer Experiment Results Analysis',
                href='/enhancer-experiment-results'
            ),
            html.Hr(),
            html.Div(
                id='current-page-context',
                style={'display': 'none'},
                children='enhancer'
            )
        ])

    def combo_aggregated_metrics_layout(df):

        columns = [{"name": i, "id": i} for i in df.columns]
        sample_columns = [col for col in df.columns if "Sample" in col]
        for col in sample_columns:
            df[col] = df[col].apply(
                lambda x: x.text_output
                if isinstance(x, MultimodalOutput) else x
            )
        sample_style = [{
            'if': {
                'column_id': col
            },
            'width': '15%'
        } for col in sample_columns]

        # styles = highlight_best_values(df, *df.columns)
        styles = generate_heatmap_style(df, *df.columns)
        styles += sample_style

        # Highlight the best_combination row
        best_combination = experiment_data.selection_output.best_combination if experiment_data.selection_output else None
        tooltip_data = []
        if best_combination:
            best_combination_str = "\n".join(
                textwrap.wrap(
                    str(best_combination).replace('"', "").replace("'", ""), 90
                )
            )
            styles.append({
                'if': {
                    'column_id':
                    'Hyperparameters',
                    'filter_query':
                    f'{{Hyperparameters}} eq "{best_combination_str}"',
                },
                'backgroundColor': '#DFF0D8',  # Light green color
                'border': '2px solid #28A745',  # Darker green border
                'color': '#155724'  # Dark text color for contrast
            })

            # If selection_reason is available, add it as a tooltip:
            if experiment_data.selection_output and experiment_data.selection_output.selection_reason:
                reason_str = ', '.join([
                    f"{k}: {v}" for k, v in
                    experiment_data.selection_output.selection_reason.items()
                ])
                tooltip_data = [{
                    'Hyperparameters': {
                        'value': reason_str,
                        'type': 'markdown'
                    }
                } if row["Hyperparameters"] == best_combination_str else {}
                                for row in df.to_dict('records')]

        evaluator_names = [
            col.replace(" Output", "") for col in df.columns if "Output" in col
        ]

        columns = [{"name": i, "id": i} for i in df.columns]
        return dash_table.DataTable(
            id='combo-metrics-table',
            columns=columns,
            data=df.to_dict('records'),
            style_data_conditional=styles,
            style_cell={
                'whiteSpace': 'normal',
                'height': '60px',
                'textAlign': 'left',
                'fontSize': 16
            },
            style_table={
                'width': '100%',
                'maxHeight': '100vh',
                'overflowY': 'auto'
            },
            style_header={
                'backgroundColor': 'rgb(230, 230, 230)',
                'fontWeight': 'bold'
            },
            style_data={
                'overflow': 'hidden',
                'textOverflow': 'ellipsis',
                'backgroundColor': 'rgb(248, 248, 248)'
            },
            style_cell_conditional=[{
                'if': {
                    'column_id': 'Hyperparameters'
                },
                'width': '40%'
            }, {
                'if': {
                    'column_id': 'Average Token Usage'
                },
                'width': '20%'
            }, {
                'if': {
                    'column_id': 'Average Latency'
                },
                'width': '20%'
            }, {
                'if': {
                    'column_id': 'Evaluator Outputs'
                },
                'width': '15%'
            }, *[{
                'if': {
                    'column_id': col + " Output"
                },
                'width': '15%'
            } for col in evaluator_names]] + sample_style,
            filter_action="native",
            sort_action="native",
            page_size=10,
            tooltip_data=tooltip_data,
            tooltip_duration=None
        )

    def image_combo_aggregated_metrics_layout(df):
        sample_columns = [col for col in df.columns if "Sample" in col]

        # Process each column with MultimodalOutput
        for col in sample_columns:
            df[col] = df[col].apply(
                lambda x: html.Div([
                    html.P(f'<text_output> {x.text_output} </text_output>'),
                    html.Div(
                        html.Img(
                            src=pil_image_to_base64(x.image_output[0]),
                            style={
                                'maxHeight': '100%',
                                'maxWidth': '100%',
                                'objectFit': 'contain'
                            }
                        ),
                        style={
                            'height': '200px',
                            'width': '200px',
                            'display': 'flex',
                            'justifyContent': 'center',
                            'alignItems': 'center',
                            'overflow': 'hidden'
                        }
                    )
                ]) if isinstance(x, MultimodalOutput) and x.image_output is
                not None else x
            )

        # Create html.Table
        table = html.Table(
            # Header
            [html.Tr([html.Th(col) for col in df.columns])] +

            # Body
            [
                html.Tr([
                    DangerouslySetInnerHTML(
                        f'<details><summary>{row[col][:250]}...</summary>{row[col]}</details>'
                    ) if col_index == 0 else html.Td(row[col])
                    for col_index, col in enumerate(df.columns)
                ]) for index, row in df.iterrows()
            ]
        )

        return table

    def video_combo_aggregated_metrics_layout(df):
        sample_columns = [col for col in df.columns if "Sample" in col]

        # Process each column with MultimodalOutput
        for col in sample_columns:
            df[col] = df[col].apply(
                lambda x: html.Div([
                    html.P(f'<text_output> {x.text_output} </text_output>'),
                    html.Div(
                        children=[
                            html.Video(
                                controls=True,
                                id='movie_player',
                                src=x.video_output[0],
                                autoPlay=False
                            )
                        ],
                        style={
                            'height': '200px',
                            'width': '200px',
                            'display': 'flex',
                            # 'justifyContent': 'center',
                            # 'alignItems': 'center',
                            # 'overflow': 'hidden'
                        }
                    )
                ]) if isinstance(x, MultimodalOutput) and x.video_output is
                not None else x
            )

        # Create html.Table
        table = html.Table(
            # Header
            [html.Tr([html.Th(col) for col in df.columns])] +

            # Body
            [
                html.Tr([
                    DangerouslySetInnerHTML(
                        f'<details><summary>{row[col][:250]}...</summary>{row[col]}</details>'
                    ) if col_index == 0 else html.Td(row[col])
                    for col_index, col in enumerate(df.columns)
                ]) for index, row in df.iterrows()
            ]
        )

        return table

    def format_with_tags(cell):
        # Split the cell content using the yival_raw_output tags
        raw_output_start = "<yival_raw_output>"
        raw_output_end = "</yival_raw_output>"

        if raw_output_start in cell and raw_output_end in cell:
            raw_output_content, rest = cell.split(raw_output_end, 1)
            raw_output_content = raw_output_content.replace(
                raw_output_start, ""
            ).strip()

            # Format the rest of the content
            formatted_evaluator_outputs = []
            for line in rest.split("\n"):
                formatted_evaluator_outputs.append(
                    "▶ " + line.strip() if ":" in line else line.strip()
                )

            # Combine raw_output and formatted evaluator outputs
            return raw_output_start + "\n" + raw_output_content + "\n" + raw_output_end + "\n" + "\n".join(
                formatted_evaluator_outputs
            )
        else:
            return cell

    def format_dataframe_column(cell):
        # If the cell contains the yival_raw_output tags
        if "<yival_raw_output>" in cell and "</yival_raw_output>" in cell:
            return format_with_tags(cell)
        else:
            # Otherwise, format the cell content using the original "▶" logic
            return "\n".join([
                "▶ " + line if ":" in line else line
                for line in cell.split("\n")
            ])

    test_data_hint = html.Div([
        html.Small(
            "Click on each test data below to perform human labeling.",
            style={
                "color": "#888888",
                "display": "block",
                "marginBottom": "10px"
            }
        )
    ])

    def group_key_combination_layout(
        group_experiment_results: List[GroupedExperimentResult],
        highlight_key: Optional[str] = None
    ):
        df_group_key = generate_group_key_combination_data(
            group_experiment_results
        )
        for col in df_group_key.columns:
            if col != "Test Data":
                df_group_key[col] = df_group_key[col].apply(
                    format_dataframe_column
                )

        csv_string = df_group_key.to_csv(index=False, encoding='utf-8')
        csv_data_url = 'data:text/csv;charset=utf-8,' + urllib.parse.quote(
            csv_string
        )

        columns = [{"name": i, "id": i} for i in df_group_key.columns]

        styles_data_conditional = [
            {
                'if': {
                    'filter_query': '{' + col + '} contains "▶"',
                },
                'backgroundColor':
                '#E6F7FF',  # Light blue color for distinction
                'paddingLeft':
                15  # Padding to make angle brackets more visible
            } for col in df_group_key.columns if col != "Test Data"
        ]
        if highlight_key:
            styles_data_conditional.append({
                'if': {
                    'column_id': 'Hyperparameters',
                    'filter_query': f'{{Hyperparameters}} eq "{highlight_key}"'
                },
                'backgroundColor': '#FFCCCC'  # Highlighting with gold color
            })
        styles_data_conditional.append({
            'if': {
                'column_id': 'Test Data'
            },
            'color': '#007BFF',  # Blue color
            'textDecoration': 'underline',
            'cursor': 'pointer'
        })
        if "Test Data" in df_group_key:
            df_group_key["Hashed Group Key"] = df_group_key["Test Data"].apply(
                lambda group_key: hashlib.sha256(group_key.encode()).hexdigest(
                )
            )
        data_dict = df_group_key.to_dict('records'),
        if include_image_base64(data_dict):
            new_data_dict = extract_and_decode_image(data_dict)
            return html.Div([
                html.A(
                    'Export to CSV',
                    id='export-link-group-key-combo',
                    download="group_key_combo.csv",
                    href=csv_data_url,
                    target="_blank"
                ),
                html.Br(), test_data_hint,
                html.Table(
                    create_table(new_data_dict), id='group-key-combo-table'
                ),
                html.Hr(),
                dcc.Link(
                    'Go back to Experiment Results Analysis',
                    href='/experiment-results'
                ),
                html.Br(),
                dcc.Link('Go to Data Analysis', href='/data-analysis'),
                html.Br()
            ])
        elif include_video(data_dict):
            new_data_dict = extract_and_decode_video(data_dict)
            return html.Div([
                html.A(
                    'Export to CSV',
                    id='export-link-group-key-combo',
                    download="group_key_combo.csv",
                    href=csv_data_url,
                    target="_blank"
                ),
                html.Br(), test_data_hint,
                html.Table(
                    create_video_table(new_data_dict),
                    id='group-key-combo-table'
                ),
                html.Hr(),
                dcc.Link(
                    'Go back to Experiment Results Analysis',
                    href='/experiment-results'
                ),
                html.Br(),
                dcc.Link('Go to Data Analysis', href='/data-analysis'),
                html.Br()
            ])
        else:
            return html.Div([
                html.A(
                    'Export to CSV',
                    id='export-link-group-key-combo',
                    download="group_key_combo.csv",
                    href=csv_data_url,
                    target="_blank"
                ),
                html.Br(),
                test_data_hint,
                dash_table.DataTable(
                    id='group-key-combo-table',
                    columns=columns,
                    data=df_group_key.to_dict('records'),
                    style_cell={
                        'whiteSpace':
                        'pre-line',  # Allows for line breaks within cells
                        'height': 'auto',
                        'textAlign': 'left',
                        'fontSize': 16,
                        'border': '1px solid #eee'
                    },
                    style_table={
                        'width': '100%',
                        'maxHeight': '100vh',
                        'overflowY': 'auto',
                        'border': '1px solid #ddd'
                    },
                    style_header={
                        'backgroundColor': 'rgb(230, 230, 230)',
                        'fontWeight': 'bold'
                    },
                    style_data={
                        'overflow': 'hidden',
                        'textOverflow': 'ellipsis',
                        'backgroundColor': 'rgb(248, 248, 248)'
                    },
                    style_data_conditional=styles_data_conditional,
                    filter_action="native",
                    sort_action="native",
                    page_size=10,
                    tooltip_duration=None
                ),
                html.Hr(),
                dcc.Link(
                    'Go back to Experiment Results Analysis',
                    href='/experiment-results'
                ),
                html.Br(),
                dcc.Link('Go to Data Analysis', href='/data-analysis'),
                html.Br()
            ])

    def determine_relative_color(value, values):
        if not isinstance(value, (int, float)):
            return "black"

        p25, p50, p75 = np.percentile(values, [25, 50, 75])

        if value <= p25:
            return "green"
        elif value <= p50:
            return "lightgreen"
        elif value <= p75:
            return "orange"
        else:
            return "red"

    def colorize_metric(combinations):
        # Extract all metrics from the results dynamically
        all_metrics = {key: [] for key in combinations[0].keys()}
        for combo in combinations:
            for key, value in combo.items():
                if isinstance(value, (int, float)):
                    all_metrics[key].append(value)

        colored_combinations = []
        for combo in combinations:
            colored_combo = {}
            for key, value in combo.items():
                color = determine_relative_color(
                    value, all_metrics.get(key, [])
                )
                colored_combo[key] = {"value": value, "color": color}
            colored_combinations.append(colored_combo)

        return colored_combinations

    def input_page_layout():
        # Function to truncate long text
        def truncate_text(text, max_length=60):  # Adjust max_length as needed
            """Truncate text to a specified length and append ellipses."""
            return text if len(text
                               ) <= max_length else text[:max_length] + "..."

        initial_combinations = {
            str(result.combination): result.combination
            for group in experiment_data.group_experiment_results
            for result in group.experiment_results
        }

        return html.Div([
            dbc.Row([
                dbc.Col(
                    [
                        dbc.Card(
                            [
                                dbc.CardHeader(
                                    html.H4("Input", className="text-center"),
                                    className="bg-light"
                                ),
                                dbc.CardBody([
                                    html.Div([
                                        dbc.Label(
                                            key,
                                            className="mr-2 font-weight-bold",
                                            width=4
                                        ),
                                        dbc.Col(
                                            dbc.Input(
                                                id=f"input-{key}",
                                                type=value,
                                                placeholder=key
                                            ),
                                            width=8
                                        )
                                    ],
                                             className=
                                             "d-flex align-items-center mb-4")
                                    for key, value in function_args.items()
                                ],
                                             className="p-4"),
                                dbc.Button(
                                    "Run",
                                    id="interactive-btn",
                                    color="primary",
                                    className="mt-2 mb-4 w-100"
                                ),
                                # Enhanced Section for Combinations Selection
                                html.Hr(),
                                html.Div([
                                    html.H5(
                                        "Select Combinations",
                                        className="text-center mt-2"
                                    ),
                                    html.P(
                                        "Choose from the available combinations to run. Multiple selections allowed.",
                                        className="text-muted small text-center"
                                    ),
                                    dcc.Dropdown(
                                        id="combinations-select",
                                        options=[{
                                            "label":
                                            truncate_text(str(combo)),
                                            "value":
                                            str(combo)
                                        } for combo in initial_combinations.
                                                 values()],
                                        multi=True,
                                        value=[
                                            str(combo) for combo in
                                            initial_combinations.values()
                                        ],
                                        style={
                                            "border": "1px solid #ced4da",
                                            "border-radius": "4px",
                                            "padding": "5px",
                                            "margin-bottom": "20px"
                                        }
                                    )
                                ],
                                         style={"padding": "10px"}),
                                # Toggle for enhancer combinations
                                dbc.Checklist(
                                    options=[{
                                        "label": "Use Enhancer Combinations",
                                        "value": "enhancer"
                                    }],
                                    value=[],
                                    id="enhancer-toggle",
                                    switch=True,
                                    inline=True,
                                    style={"padding": "10px"}
                                )
                            ],
                            className="m-4 shadow-sm rounded"
                        ),
                    ],
                    width=3
                ),
                dbc.Col([
                    dbc.Card([
                        dbc.CardHeader(
                            html.H4("Results", className="text-center"),
                            className="bg-light"
                        ),
                        dcc.Loading(
                            id="loading-results",
                            type="default",
                            children=html.
                            Div(id="results-section", className="p-4")
                        )
                    ],
                             className="m-4 shadow-sm rounded")
                ],
                        width=9)
            ]),
            dbc.Container(fluid=True, className="p-3")
        ])

    def use_best_result_layout():
        if experiment_data.enhancer_output:
            best_combination = experiment_data.enhancer_output.selection_output.best_combination
        elif experiment_data.selection_output:
            best_combination = experiment_data.selection_output.best_combination
        else:
            best_combination = experiment_data.group_experiment_results[
                0].experiment_results[0].combination
        best_combination = str(best_combination)

        return html.Div([
            dbc.Row([
                dbc.Col(
                    [
                        dbc.Card(
                            [
                                dbc.CardHeader(
                                    html.
                                    H4("Parameters", className="text-center"),
                                    className="bg-light"
                                ),
                                dbc.CardBody([
                                    html.Div([
                                        dbc.Label(
                                            key,
                                            className="mr-2 font-weight-bold",
                                            width=4
                                        ),
                                        dbc.Col(
                                            dbc.Input(
                                                id=f"input-{key}",
                                                type=value,
                                                placeholder=key
                                            ),
                                            width=8
                                        )
                                    ],
                                             className=
                                             "d-flex align-items-center mb-4")
                                    for key, value in
                                    list(function_args.items())[:-1]
                                ],
                                             className="p-4"),
                                dbc.Button(
                                    "Run",
                                    id="best-result-btn",
                                    color="primary",
                                    className="mt-2 mb-4 w-100"
                                ),
                                # Enhanced Section for Combinations Selection
                                html.Hr(),
                                html.Div([
                                    html.H5(
                                        "Best Combination",
                                        className="text-center mt-2"
                                    ),
                                    html.P(
                                        "This is the best combination from the available combinations.",
                                        className="text-muted small text-center"
                                    ),
                                    html.P(
                                        best_combination,
                                        id="best-combination",
                                        style={
                                            "border": "1px solid #ced4da",
                                            "border-radius": "4px",
                                            "padding": "5px",
                                            "margin-bottom": "20px"
                                        }
                                    )
                                ],
                                         style={"padding": "10px"}),
                            ],
                            className="m-4 shadow-sm rounded"
                        ),
                    ],
                    width=3
                ),
                dbc.Col([
                    dbc.Card([
                        dbc.CardHeader(
                            html.H4("Results", className="text-center"),
                            className="bg-light"
                        ),
                        dcc.Loading(
                            id="loading-results",
                            type="default",
                            children=html.
                            Div(id="best-results-section", className="p-4")
                        )
                    ],
                             className="m-4 shadow-sm rounded")
                ],
                        width=9)
            ]),
            dbc.Container(fluid=True, className="p-3")
        ])

    def display_group_experiment_result_layout(
        hashed_group_key, experiment_config, is_from_enhancer=False
    ):
        group_result = get_group_experiment_result_from_hash(hashed_group_key)
        if not group_result:
            return html.Div(
                "No data found for this group key.",
                style={
                    'font-size': '24px',
                    'color': 'red',
                    'margin': '40px',
                    'text-align': 'center'
                }
            )
        initial_slider_data = {}
        sanitized_key = sanitize_group_key(group_result.group_key)

        children = [
            html.H3(
                f"Group Key: {sanitized_key}",
                style={
                    'text-align': 'center',
                    'margin-bottom': '20px',
                    'color': '#5D6D7E'
                }
            ),
            html.H2(
                "Rate experiment results",
                style={
                    'text-align': 'center',
                    'margin-bottom': '70px',
                    'font-weight': 'bold',
                    'color': '#2E86C1',
                    'border-bottom': '4px solid #3498DB',
                    'padding-bottom': '30px',
                    'font-size': '28px'
                }
            ),
            dcc.Input(
                id='current-group-key', type='hidden', value=hashed_group_key
            )
        ]
        children.append(
            dcc.Store(id='is-from-enhancer', data=is_from_enhancer)
        )

        for index, exp_result in enumerate(group_result.experiment_results):
            # Displaying only the raw output
            children.append(
                html.Div(
                    "Raw Output:",
                    style={
                        'font-weight': 'bold',
                        'font-size': '26px',
                        'margin-top': '50px',
                        'color': '#2C3E50'
                    }
                )
            )
            if exp_result.raw_output.image_output:
                images = []
                for image in exp_result.raw_output.image_output:
                    img_str = pil_image_to_base64(image)
                    img = html.Img(
                        src=img_str,
                        style={
                            'width': '200px',
                            'height': '200px',
                            'margin': 'auto'
                        }
                    )
                    images.append(img)
                text = html.P(
                    exp_result.raw_output.text_output,
                    style={'grid-column': 'span 4'
                           }  # Make the text span 4 columns
                ) if exp_result.raw_output.text_output else None
                content = html.Div([text] + images,
                                   style={
                                       'display': 'grid',
                                       'grid-template-columns':
                                       'repeat(4, 1fr)',
                                       'justify-content': 'center',
                                       'align-items': 'center'
                                   })
            elif exp_result.raw_output.video_output:
                videos = []
                for video in exp_result.raw_output.video_output:
                    vid = html.Video(
                        controls=True,
                        id='movie_player',
                        src=video,
                        autoPlay=False,
                        style={
                            'width': '200px',
                            'height': '200px',
                            'margin': 'auto'
                        }
                    )
                    videos.append(vid)
                text = html.P(
                    exp_result.raw_output.text_output,
                    style={'grid-column': 'span 4'
                           }  # Make the text span 4 columns
                ) if exp_result.raw_output.text_output else None
                content = html.Div([text] + videos,
                                   style={
                                       'display': 'grid',
                                       'grid-template-columns':
                                       'repeat(4, 1fr)',
                                       'justify-content': 'center',
                                       'align-items': 'center'
                                   })
            else:
                content = str(exp_result.raw_output.text_output)

            children.append(
                html.Div(
                    content,
                    style={
                        'margin': '25px 0',
                        'border': '2px solid #AED6F1',
                        'padding': '25px',
                        'background-color': '#EAF2F8',
                        'border-radius': '10px',
                        'box-shadow': '0 4px 12px rgba(0, 0, 0, 0.1)',
                        'font-size': '22px'
                    }
                )
            )

            for rating_config_index, rating_config in enumerate(
                experiment_config["human_rating_configs"] or []
            ):
                existing_human_evaluator = next((
                    e for e in exp_result.evaluator_outputs
                    if e.name == "human_evaluator"
                    and e.display_name == rating_config['name']
                ), None)

                # Displaying rating instructions if present in rating_config
                if "instructions" in rating_config:
                    children.append(
                        html.Div(
                            "Instructions:",
                            style={
                                'font-weight': 'bold',
                                'font-size': '24px',
                                'margin-top': '40px',
                                'color': '#5D6D7E'
                            }
                        )
                    )
                    children.append(
                        html.Div(
                            rating_config["instructions"],
                            style={
                                'margin': '20px 0',
                                'background-color': '#D5F5E3',
                                'padding': '20px',
                                'border-radius': '10px',
                                'font-size': '22px'
                            }
                        )
                    )

                # Slider to represent the rating scale
                slider_id = {
                    "type": "rating-slider",
                    "index": f"{rating_config['name']}-{index}"
                }
                slider_key = f"{rating_config['name']}-{index}"

                # Setting default value if evaluator output exists
                if existing_human_evaluator:
                    default_value = existing_human_evaluator.result if existing_human_evaluator else (
                        rating_config["scale"][0] + rating_config["scale"][1]
                    ) / 2
                    initial_slider_data[slider_key] = default_value
                else:
                    default_value = None

                children.append(
                    html.Div(
                        rating_config["name"],
                        style={
                            'font-size': '24px',
                            'margin-top': '40px',
                            'color': '#5D6D7E'
                        }
                    )
                )
                scale_bar = dcc.Slider(
                    min=rating_config["scale"][0],
                    max=rating_config["scale"][1],
                    id=slider_id,
                    marks={
                        i: str(i)
                        for i in range(
                            int(rating_config["scale"][0]),
                            int(rating_config["scale"][1]) + 1
                        )
                    },
                    value=default_value if default_value else
                    (rating_config["scale"][0] + rating_config["scale"][1]) /
                    2,
                    disabled=False
                )
                children.append(scale_bar)
        children.append(
            html.Button(
                'Save',
                id='save-button',
                style={
                    'display': 'block',
                    'margin': '40px auto',
                    'padding': '10px 20px',
                    'background-color': '#2E86C1',
                    'color': 'white',
                    'border': 'none',
                    'border-radius': '5px',
                    'cursor': 'pointer'
                }
            )
        )

        # Add the output-div here
        children.append(html.Div(id='output-div'))
        children.append(
            dcc.Store(id='slider-values-store', data=initial_slider_data)
        )

        return html.Div(
            children,
            style={
                'padding': '80px',
                'background-color': '#FCF3CF',
                'border-radius': '20px',
                'box-shadow': '0 8px 25px rgba(0, 0, 0, 0.1)',
                'font-family': 'Arial, sans-serif',
                'width': '90%',
                'margin': '2% auto'
            }
        )

    app = dash.Dash(__name__, external_stylesheets=[dbc.themes.FLATLY])

    app.config.suppress_callback_exceptions = True

    df = generate_combo_metrics_data(
        experiment_data.combination_aggregated_metrics,
        experiment_data.group_experiment_results
    )

    @app.callback(
        dash.dependencies.Output('page-content', 'children'),
        [dash.dependencies.Input('url', 'pathname')]
    )
    def display_page(pathname):
        if pathname.startswith('/rating-result/'):
            hashed_group_key = pathname.split('/')[-1]
            is_from_enhancer = "?source=enhancer" in hashed_group_key
            hashed_group_key = hashed_group_key.replace('?source=enhancer', '')
            return display_group_experiment_result_layout(
                hashed_group_key, experiment_config, is_from_enhancer
            )
        elif pathname == '/experiment-results':
            return experiment_results_layout()
        elif pathname == '/group-key-combo':
            return combo_page_layout()
        elif pathname == '/enhancer-experiment-results':
            return enhancer_experiment_results_layout()
        elif pathname == '/enhancer-group-key-combo':
            return enhancer_combo_page_layout()
        elif pathname == '/interactive':
            return input_page_layout()
        elif pathname == '/use-best-result':
            return use_best_result_layout()
        elif pathname == '/data-analysis':
            return data_analysis_layout()
        else:
            return index_page()

    @app.callback(
        Output('comparative-scatter-plot-token', 'figure'),
        [Input('evaluator-dropdown-token', 'value')]
    )
    def update_comparative_scatter_token(selected_evaluator):
        if selected_evaluator is None:
            fig = px.scatter(title="No selected_evaluator")
            return fig
        return px.scatter(
            df,
            x='Average Token Usage',
            y=selected_evaluator,
            title=
            f'Comparative Scatter plot of Average Token Usage vs {selected_evaluator}',
            size_max=15
        )

    @app.callback(
        Output('comparative-scatter-plot-latency', 'figure'),
        [Input('evaluator-dropdown-latency', 'value')]
    )
    def update_comparative_scatter_latency(selected_evaluator):
        if selected_evaluator is None:
            fig = px.scatter(title="No selected_evaluator")
            return fig
        return px.scatter(
            df,
            x='Average Latency',
            y=selected_evaluator,
            title=
            f'Comparative Scatter plot of Average Latency vs {selected_evaluator}',
            size_max=15
        )

    @app.callback(
        Output('hexagram-container', 'children'),
        [Input('cast-fortune-btn', 'n_clicks')]
    )
    def update_hexagram(n_clicks):
        import random
        hexagram = random.choice(HEXAGRAMS)
        return [
            generate_hexagram_figure(hexagram["figure"]),
            html.H4(hexagram["name"]),
            html.P(hexagram["description"]),
            html.P(
                hexagram["reading"],
                style={
                    "color": "blue",
                    "fontStyle": "italic"
                }
            )
        ]

    @app.callback(
        Output('correlation-coefficient', 'children'),
        [Input('evaluator-dropdown-token', 'value')]
    )
    def display_correlation_coefficient(selected_evaluator):
        try:
            # Create a temporary variable to hold numerical data
            if df[selected_evaluator
                  ].dtype == 'object':  # if it's a string type
                # Extract the numerical part from each entry
                temp_evaluator_data = df[selected_evaluator].str.extract(
                    '(\d+\.\d+)'
                ).astype(float)
            else:
                # If it's already numerical, use it as is
                temp_evaluator_data = df[selected_evaluator]

            # Ensure 'Average Token Usage' is also numerical
            temp_avg_token_usage = pd.to_numeric(
                df['Average Token Usage'], errors='coerce'
            )

            # Check if the columns exist and are not null, and if their lengths match
            if temp_avg_token_usage is not None and temp_evaluator_data is not None and len(
                temp_avg_token_usage
            ) == len(temp_evaluator_data):
                correlation = temp_avg_token_usage.corr(temp_evaluator_data)
                return f"Correlation between Average Token Usage and {selected_evaluator}: {correlation:.2f}"
            else:
                return "Data not available for correlation calculation."

        except Exception as e:
            return f"An error occurred: {str(e)}"

    @app.callback(
        Output('slider-values-store', 'data'),
        [Input({
            'type': 'rating-slider',
            'index': ALL
        }, 'value')], [State('slider-values-store', 'data')],
        prevent_initial_call=True
    )
    def update_slider_store(slider_values, current_data):
        # The IDs of triggered inputs
        ctx = dash.callback_context

        triggered_slider_ids = [
            json.loads(t['prop_id'].split('.')[0])['index']
            for t in ctx.triggered
        ]
        all_slider_ids = [
            json.loads(prop_id.split('.')[0])['index']
            for prop_id in ctx.inputs.keys()
        ]

        # Create a dictionary to map slider IDs to their values
        slider_id_to_value = dict(zip(all_slider_ids, slider_values))

        for slider_id in triggered_slider_ids:
            current_data[slider_id] = slider_id_to_value[slider_id]

        return current_data

    @app.callback(
        Output('output-div', 'children'), [
            Input('save-button', 'n_clicks'),
            State('current-group-key', 'value'),
            State('slider-values-store', 'data'),
            State('is-from-enhancer', 'data')
        ]
    )  # We'll handle the dynamic State components inside the function itself
    def update_output(
        n_clicks, hashed_group_key, slider_values_store, is_from_enhancer
    ):
        if not n_clicks:
            return dash.no_update

        # 2. Determine the Current Group
        group_result = get_group_experiment_result_from_hash(hashed_group_key)
        if not group_result:
            return "Error fetching group result."

        for index, exp_result in enumerate(group_result.experiment_results):
            for rating_config in experiment_config["human_rating_configs"]:
                slider_key = f"{rating_config['name']}-{index}"
                if slider_key in slider_values_store:
                    slider_value = slider_values_store[slider_key]
                    existing_output = next((
                        e for e in exp_result.evaluator_outputs
                        if e.name == "human_evaluator"
                        and e.display_name == rating_config['name']
                    ), None)
                    if existing_output:
                        existing_output.result = slider_value
                    else:
                        new_output = EvaluatorOutput(
                            name="human_evaluator",
                            result=slider_value,
                            display_name=rating_config['name'],
                            metric_calculators=[{
                                'method': 'AVERAGE'
                            }]
                        )
                        exp_result.evaluator_outputs.append(new_output)

        if is_from_enhancer:
            results = []
            for e in experiment_data.enhancer_output.group_experiment_results:
                for r in e.experiment_results:
                    results.append(r)
            r = generate_experiment(
                results=results,
                evaluator=None,
                evaluate_all=False,
                evaluate_group=False
            )
            experiment_data.enhancer_output.combination_aggregated_metrics = r.combination_aggregated_metrics
        else:
            results = []
            for e in experiment_data.group_experiment_results:
                for r in e.experiment_results:
                    results.append(r)
            r = generate_experiment(
                results=results,
                evaluator=None,
                evaluate_all=False,
                evaluate_group=False
            )
            experiment_data.combination_aggregated_metrics = r.combination_aggregated_metrics
        return "Data saved successfully!"

    @app.callback(
        dash.dependencies.Output('url', 'pathname'), [
            dash.dependencies.Input('group-key-combo-table', 'active_cell'),
            dash.dependencies.Input('group-key-combo-table', 'data'),
            dash.dependencies.Input('current-page-context', 'children')
        ]
    )
    def navigate_to_hashed_page(active_cell, table_data, page_context):
        if page_context is None or page_context == dash.no_update:
            page_context = 'default'
        if active_cell:
            row = active_cell["row"]
            col_id = active_cell["column_id"]
            if col_id == "Test Data":
                hashed_group_key = table_data[row]["Hashed Group Key"]
                if page_context == 'enhancer':
                    return f'/rating-result/{hashed_group_key}?source=enhancer'
                return f'/rating-result/{hashed_group_key}'
        return dash.no_update

    @app.callback(
        Output("export-data-modal", "is_open"),
        [
            Input("export-btn", "n_clicks"),
            Input("close-export-modal", "n_clicks"),
            Input("confirm-export", "n_clicks")
        ],
        [State("export-data-modal", "is_open")],
    )
    def toggle_modal(n1, n2, n3, is_open):
        if n1 or n2 or n3:
            return not is_open
        return is_open

    @app.callback(
        Output('button-click-output', 'children'),
        Input('confirm-export', 'n_clicks'), State('file-path-input', 'value')
    )
    def export_callback(n, file_path):
        if n:
            import pickle
            with open(file_path, "wb") as f:
                pickle.dump(experiment_data, f)
            return f"Data would be exported to: {file_path}"
        return dash.no_update

    all_results: List[html.Div] = []

    @app.callback(
        Output("results-section", "children"),
        Input("interactive-btn", "n_clicks"),
        [State(f"input-{key}", "value") for key in function_args.keys()] + [
            State("combinations-select", "value"),
            State("enhancer-toggle", "value")
        ],
        prevent_initial_call=True
    )
    def update_results(n_clicks, *input_values_and_combinations_and_toggle):
        if not n_clicks:
            return []
        *input_values, selected_combinations_str, use_enhancer = input_values_and_combinations_and_toggle
        use_enhancer = "enhancer" in use_enhancer if use_enhancer else False
        if selected_combinations_str is None:
            return html.Div(
                "Please select at least one combination.",
                style={"color": "red"}
            )

        selected_combinations = [
            ast.literal_eval(combo_str)
            for combo_str in selected_combinations_str
        ]

        missing_fields = [
            key for i, key in enumerate(function_args.keys())
            if input_values[i] is None
            and key != "yival_expected_result (Optional)"
        ]
        if missing_fields:
            return html.Div(
                f"Please fill out the following required fields: {', '.join(missing_fields)}",
                style={"color": "red"}
            )

        content = {}
        i = 0
        expected_result = None
        for k, v in function_args.items():
            if k == "yival_expected_result (Optional)" and input_values[
                i] is not None:
                expected_result = input_values[i]
            else:
                if input_values[i] is not None:
                    content[k] = input_values[i]
            i += 1
        input_data = InputData(
            content=content, expected_result=expected_result
        )
        results: List[ExperimentResult] = []
        with ThreadPoolExecutor() as executor:
            for res in executor.map(
                parallel_task, [input_data], [selected_combinations], [logger],
                [evaluator]
            ):
                results.extend(res)

        if interactive_mode:
            ress = []
            for g in experiment_data.group_experiment_results:
                for r in g.experiment_results:
                    ress.append(r)
            ress.extend(results)
            updated_experiment = generate_experiment(ress, evaluator)
            experiment_data.group_experiment_results = updated_experiment.group_experiment_results
            experiment_data.combination_aggregated_metrics = updated_experiment.combination_aggregated_metrics
        all_metrics = {
            "Latency": [result.latency for result in results],
            "Token Usage": [result.token_usage for result in results]
        }

        # Extracting numeric evaluator outputs
        for result in results:
            for output in result.evaluator_outputs:
                if isinstance(output.result, (int, float)):
                    if output.name not in all_metrics:
                        all_metrics[output.name] = []
                    all_metrics[output.name].append(output.result)

        current_result = [
            html.Div([
                html.H5(f"Result {i+1}"),
                html.Ul([
                    html.Li(f"Combination: {result.combination}"),
                    html.Li("Text Raw Output:"),
                    html.Div(
                        handle_output(result.raw_output.text_output),
                        className="raw-output"
                    ),
                    html.Li("Image Raw Output:"),
                    html.Div(
                        handle_output(result.raw_output.image_output),
                        className="raw-output"
                    ),
                    html.Li(
                        f"Latency: {result.latency} s",
                        style={
                            "color":
                            determine_relative_color(
                                result.latency, all_metrics["Latency"]
                            )
                        }
                    ),
                    html.Li(
                        f"Token Usage: {result.token_usage}",
                        style={
                            "color":
                            determine_relative_color(
                                result.token_usage, all_metrics["Token Usage"]
                            )
                        }
                    ),
                    html.Li("Evaluator Outputs:"),
                    html.Ul([
                        html.Li(
                            f"{output.name} ({output.display_name}): {output.result}",
                            style={
                                "color":
                                determine_relative_color(
                                    output.result,
                                    all_metrics.get(output.name, [])
                                ) if isinstance(output.result,
                                                (int, float)) else "black"
                            }
                        ) for output in result.evaluator_outputs
                    ])
                ])
            ]) for i, result in enumerate(results)
        ]

        input_summary = ", ".join([
            f"{key}: {value}"
            for key, value in zip(function_args.keys(), input_values)
        ])
        toggle_id = {"type": "toggle", "index": n_clicks}
        collapse_id = {"type": "collapse", "index": n_clicks}

        current_result_card = html.Div([
            dbc.Button(
                input_summary,
                id=toggle_id,
                className="mb-2 w-100",
                color="info"
            ),
            dbc.Collapse([
                item for sublist in [[res, html.Hr()]
                                     for res in current_result]
                for item in sublist
            ],
                         id=collapse_id,
                         is_open=True)
        ],
                                       className="mb-3")

        all_results.insert(0, current_result_card)

        return all_results

    @app.callback(
        Output("best-results-section", "children"),
        Input("best-result-btn", "n_clicks"),
        [
            State(f"input-{key}", "value")
            for key in list(function_args.keys())[:-1]
        ] + [State("best-combination", "children")],
        prevent_initial_call=True
    )
    def update_best_results(n_clicks, *input_values_and_combinations):
        if not n_clicks:
            return []
        *input_values, best_combination = input_values_and_combinations
        # Create a new InputData instance
        input_data = InputData(
            dict(zip(list(function_args.keys())[:-1], input_values))
        )
        missing_fields = [
            key for i, key in enumerate(list(function_args.keys())[:-1])
            if input_values[i] is None
            and key != "yival_expected_result (Optional)"
        ]
        if missing_fields:
            return html.Div(
                f"Please fill out the following required fields: {', '.join(missing_fields)}",
                style={"color": "red"}
            )

        state.active = True
        best_varation = json.loads(best_combination)
        for key in best_varation:
            best_varation[key] = [best_varation[key]]
        state.current_variations = best_varation
        res = call_function_from_string(
            experiment_config["custom_function"],  # type: ignore
            **input_data.content,
            state=state
        ) if "custom_function" in experiment_config else None  #type: ignore

        current_result = [
            html.Div([
                html.Ul([
                    html.Li("Text Raw Output:"),
                    html.Div(
                        handle_output(res.text_output), className="raw-output"
                    ),
                ] + ([
                    html.Li("Image Raw Output:"),
                    html.Div(
                        handle_output(res.image_output),
                        className="raw-output"
                    )
                ] if res.image_output is not None else []))
            ])
        ]

        input_summary = ", ".join([
            f"{key}: {value}" for key, value in
            zip(list(function_args.keys())[:-1], input_values)
        ])
        toggle_id = {"type": "toggle", "index": n_clicks}
        collapse_id = {"type": "collapse", "index": n_clicks}
        current_result_card = html.Div([
            dbc.Button(
                input_summary,
                id=toggle_id,
                className="mb-2 w-100",
                color="info"
            ),
            dbc.Collapse([
                item for sublist in [[res, html.Hr()]
                                     for res in current_result]
                for item in sublist
            ],
                         id=collapse_id,
                         is_open=True)
        ],
                                       className="mb-3")

        all_results.insert(0, current_result_card)
        return all_results

    def truncate_text(text, max_length=60):  # Adjust max_length as needed
        """Truncate text to a specified length and append ellipses."""
        return text if len(text) <= max_length else text[:max_length] + "..."

    @app.callback(
        Output("combinations-select", "options"),
        [Input("enhancer-toggle", "value")]
    )
    def update_combinations_options(use_enhancer):
        unique_combination = {}

        if "enhancer" in use_enhancer:
            # Use enhancer combinations
            for group in experiment_data.enhancer_output.group_experiment_results:
                for result in group.experiment_results:
                    if str(result.combination) not in unique_combination:
                        unique_combination[str(result.combination)
                                           ] = result.combination

        else:
            for group in experiment_data.group_experiment_results:
                for result in group.experiment_results:
                    if str(result.combination) not in unique_combination:
                        unique_combination[str(result.combination)
                                           ] = result.combination

            # Combine with all_combinations
            for combination in all_combinations:
                if str(combination) not in unique_combination:
                    unique_combination[str(combination)] = combination

        return [{
            "label": truncate_text(str(combo)),
            "value": str(combo)
        } for combo in unique_combination.values()]

    @app.callback(
        Output({
            "type": "collapse",
            "index": MATCH
        }, "is_open"),
        Input({
            "type": "toggle",
            "index": MATCH
        }, "n_clicks"),
        State({
            "type": "collapse",
            "index": MATCH
        }, "is_open"),
        prevent_initial_call=True
    )
    def toggle_collapse(n, is_open):
        return not is_open

    app.layout = html.Div(
        [
            html.Div(
                [
                    dcc.Location(id='url', refresh=False),
                    generate_navigation(),
                    html.Div(
                        id='page-content',
                        style={
                            'fontFamily': 'Arial, sans-serif',
                            'margin': '2% 10%',
                            'padding': '2% 3%',
                            'border': '1px solid #ddd',
                            'borderRadius': '5px',
                            'backgroundColor': '#f9f9f9',
                            'fontSize': '18px',
                            'boxShadow': '0 4px 8px 0 rgba(0, 0, 0, 0.1)',
                        }
                    )
                ],
                className="main-content",
                style={
                    'backgroundColor': '#f9f9f9',  # Light Grey
                    'padding': '20px'
                }
            ),
            dbc.Modal([
                dbc.ModalHeader("Export Data"),
                dbc.ModalBody([
                    html.
                    Label("Please enter the file path for exporting data:"),
                    dcc.Input(
                        id="file-path-input",
                        type="text",
                        placeholder="Enter file path..."
                    ),
                ]),
                dbc.ModalFooter([
                    dbc.Button(
                        "Close", id="close-export-modal", className="ml-auto"
                    ),
                    dbc.
                    Button("Export", id="confirm-export", className="ml-auto"),
                ]),
            ],
                      id="export-data-modal"),
            html.Div(id='button-click-output')
        ],
        className="main-wrapper"
    )

    return app


def display_results_dash(
    experiment_data: Experiment,
    experiment_config,
    all_combinations,
    state,
    logger,
    evaluator,
    interactive=False,
    port=8074
):
    if experiment_data.enhancer_output:
        for group_result in experiment_data.enhancer_output.group_experiment_results:
            experiment_results = []
            seen = set()
            for r in group_result.experiment_results:
                if str(r.combination) in seen:
                    continue
                else:
                    seen.add(str(r.combination))
                    experiment_results.append(r)
            group_result.experiment_results = experiment_results
    function_args = get_function_args(experiment_config["custom_function"])
    function_args["yival_expected_result (Optional)"] = 'str'
    app = create_dash_app(
        experiment_data, experiment_config, function_args, all_combinations,
        state, logger, evaluator, interactive
    )
    if os.environ.get("ngrok", False):
        public_url = ngrok.connect(port)
        print(f"Access Yival from this public URL :{public_url}")
        app.run(debug=False, port=port)
    else:
        app.run(debug=False, port=port)



"""
This module is the back up trainer

It will only be called when the dependency is not imported correctly
"""

from ..schemas.experiment_config import Experiment, ExperimentConfig, TrainerOutput
from .base_trainer import BaseTrainer


class BackUpTrainer(BaseTrainer):

    def train(
        self,
        experiment: Experiment,
        config: ExperimentConfig,
    ) -> TrainerOutput:
        raise ImportError(
            "Trainer was not successfully imported. Please check your dependencies."
        )


"""
Normalization functions

"""
import numpy as np


def min_max_normalization(matrix: np.ndarray) -> np.ndarray:
    """normalize matrix in min_max method"""
    matrix_min = matrix.min(axis=0)
    matrix_max = matrix.max(axis=0)
    normalized_matrix = (matrix -
                         matrix_min) / (matrix_max - matrix_min + 1e-10)
    return normalized_matrix


def z_score_normalizatioin(matrix: np.ndarray) -> np.ndarray:
    """normalize matrix in z_score method"""
    matrix_mean = matrix.mean(axis=0)
    matrix_std = matrix.std(axis=0)
    normalized_matrix = (matrix - matrix_mean) / (matrix_std + 1e-10)
    return normalized_matrix


NORMALIZATION_FUNCS = {
    'min-max': min_max_normalization,
    'z-score': z_score_normalizatioin,
}


"""
Experiment State Module.

This module defines the `ExperimentState` class to manage the state of active
experiments and their variations. The state facilitates the retrieval and
management of variations associated with different experiments, providing a
mechanism to cycle through the variations and track the experiment's state.
"""

import copy
import threading
from collections import defaultdict
from itertools import product
from typing import Any, Dict, Iterator, List, Optional, Union

from ..schemas.experiment_config import ExperimentConfig
from ..schemas.varation_generator_configs import BaseVariationGeneratorConfig
from ..variation_generators.base_variation_generator import BaseVariationGenerator


class ExperimentState:
    """
    Represents the state for managing experiment variations.

    This class maintains the state of active experiments and their variations.

    Attributes:
        active (bool): Indicates if the experiment is currently active.
        current_variations (Dict[str, List[Any]]): A dictionary where keys are
        experiment names and values are lists of variations.
        counters (Dict[str, int]): A counter for each experiment name to
        rotate through its variations.

    Methods:
        get_next_variation(name: str) -> Optional[Any]:
            Depending on the global ExperimentState's activity status,
            retrieves the next variation for the associated experiment name.
            If the state is inactive or no variations are found, returns None.
    """

    _default_state = None
    _thread_local_state = threading.local()

    @staticmethod
    def get_instance():
        if not ExperimentState._default_state:
            ExperimentState.get_default_state()
        if not hasattr(ExperimentState._thread_local_state, "_instance"):
            ExperimentState._thread_local_state._instance = copy.deepcopy(
                ExperimentState._default_state
            )
        return ExperimentState._thread_local_state._instance

    @staticmethod
    def get_default_state():
        if not ExperimentState._default_state:
            ExperimentState._default_state = ExperimentState()
        return ExperimentState._default_state

    def __init__(self) -> None:
        self.active: bool = False
        self.current_variations: Dict[str, List[Any]] = {}
        self.counters: Dict[str, int] = defaultdict(int)
        self.config: Optional[ExperimentConfig] = None

    def get_next_variation(self, name: str) -> Optional[Any]:
        variations = self.current_variations.get(name, [])
        if self.counters[name] < len(variations):
            variation = variations[self.counters[name]]
            self.counters[name] += 1
            return variation
        return None

    def get_all_variation_combinations(self) -> List[Dict[str, Any]]:
        """
        Returns a list of dictionaries where each dictionary 
        represents a unique combination of variations.
        """
        all_variations = []
        for name, variations in self.current_variations.items():
            all_variations.append([(name, variation)
                                   for variation in variations])
        combinations = []
        for combo in product(*all_variations):
            combo_dict = {name: variation for name, variation in combo}
            combinations.append(combo_dict)
        return combinations

    def initialize_variations_from_config(self) -> None:
        """
        Initializes the experiment variations using the provided
        ExperimentConfig.
        """
        if self.config and self.config.variations:
            for wrapper_config in self.config.variations:
                if not isinstance(wrapper_config, dict):
                    wrapper_config = wrapper_config.asdict()  # type: ignore
                if "variations" in wrapper_config:  # type: ignore
                    variations = [
                        var_variation["instantiated_value"] for var_variation
                        in wrapper_config["variations"]  # type: ignore
                    ]
                    self.set_variations_for_experiment(
                        wrapper_config["name"],  # type: ignore
                        variations
                    )
                if "generator_name" in wrapper_config:  # type: ignore
                    generator_cls = BaseVariationGenerator.get_variation_generator(
                        wrapper_config["generator_name"]  # type: ignore
                    )
                    config_cls = BaseVariationGenerator.get_config_class(
                        wrapper_config["generator_name"]  # type: ignore
                    )
                    if generator_cls:
                        if config_cls:
                            if "generator_config" in wrapper_config:  # type: ignore
                                if isinstance(
                                    wrapper_config["generator_config"
                                                   ],  # type: ignore
                                    dict
                                ):
                                    config_data = wrapper_config[  # type: ignore
                                        "generator_config"]
                                else:
                                    config_data = wrapper_config[  # type: ignore
                                        "generator_config"].asdict()
                                config_instance = config_cls(**config_data)
                            else:
                                config_instance = config_cls()
                            generator_instance = generator_cls(config_instance)
                        else:
                            generator_instance = generator_cls(
                                BaseVariationGeneratorConfig()
                            )
                        vs = []
                        for vars in generator_instance.generate_variations():
                            for var in vars:
                                vs.append(var.instantiated_value)
                        self.set_variations_for_experiment(
                            wrapper_config["name"],  # type: ignore
                            vs
                        )

    def set_variations_for_experiment(
        self, name: str, variations: Union[List[Any], Iterator[Any]]
    ) -> None:
        existing_variations = self.current_variations.get(name, [])
        existing_variations.extend(variations)
        self.current_variations[name] = existing_variations

    def clear_variations_for_experiment(self) -> None:
        self.current_variations.clear()

    def set_experiment_config(self, config: Any) -> None:
        if isinstance(config, dict):
            self.config = ExperimentConfig(**config)
        else:
            self.config = config
        self.initialize_variations_from_config()

    def set_specific_variation(self, name: str, variation: Any) -> None:
        """
        Sets a specific variation for an experiment without cycling through
        the variations.
        """
        self.current_variations[name] = [variation]
        self.counters[name] = 0



import json
import os
import pickle
import time
from typing import Dict, List

import replicate
import requests

from yival.dataset.data_utils import (
    evaluate_condition,
    read_code_from_path_or_module,
    transform_experiment_result_generic,
)
from yival.schemas.experiment_config import Experiment, ExperimentResult


def _format_data_for_replicate_finetune(data: list[dict[str, str]]):
    formatted_data = []

    for entry in data:
        chat_entry: dict[str, str] = {}
        prompt = ""
        if "Instruction" in entry:
            prompt += entry["Instruction"] + "\n"

        prompt += entry["Input"]
        chat_entry.update({"prompt": prompt, "completion": entry["Output"]})

        formatted_data.append(chat_entry)

    return formatted_data


def _extract_from_input_data(result: Experiment) -> List[Dict[str, str]]:
    result_pairs = []
    for rs in result.group_experiment_results:
        input_data = json.loads(rs.group_key)
        finetune_data = {
            "prompt":
            "Translate the following english to chinese" + "\n" +
            input_data['content']['teacher_quiz'],
            "completion":
            input_data['expected_result']
        }
        result_pairs.append(finetune_data)

    print(f"[INFO] len format data: {len(result_pairs)}")
    return result_pairs


def finetune(
    input_file: str,
    condition: str,
    custom_function: str,
    destination: str,
    model_name: str,
    num_train_epochs: int = 3,
    support_expected_value: bool = False,
    system_prompt: str = "",
) -> str:
    """
    Fine-tunes a replicate model using provided data and conditions.

    Args:
    - input_file (str): Path to the input file containing experiment results.
    - condition (str): Condition to evaluate for extracting relevant results.
    - custom_function (str): Path or module containing the custom function used in the experiment.
    - destination: (str): The model to push the trained version to .
    - model_name: (str): Model name that will be used for finetune.
    - num_train_epochs: (int, optional): Number of epochs to train the model.
    - system_prompt (str, optional): System message to prepend to each chat. Defaults to None.

    Returns:
    - str: ID of the fine-tuned model.
    """

    code = read_code_from_path_or_module(custom_function)
    if not code:
        print("Failed to read code from path or module.")
        return ""

    with open(input_file, 'rb') as f:
        result: Experiment = pickle.load(f)

    if support_expected_value:
        formatted_data = _extract_from_input_data(result)
    else:
        result_pairs = []
        for combo_result in result.combination_aggregated_metrics:
            results: List[ExperimentResult] = combo_result.experiment_results
            for rs in results:
                if rs.evaluator_outputs:
                    for evaluator_output in rs.evaluator_outputs:
                        condition_met = evaluate_condition(
                            condition, evaluator_output
                        )
                        if condition_met:
                            result_pair = transform_experiment_result_generic(
                                code, rs
                            )
                            result_pairs.append(result_pair)

        formatted_data = _format_data_for_replicate_finetune(result_pairs)

    headers = {"Authorization": f"Token {os.environ['REPLICATE_API_TOKEN']}"}

    with open("data.jsonl", "w", encoding='utf-8') as f:
        for entry in formatted_data:
            json_str = json.dumps(entry, separators=(',', ':'))
            f.write(json_str + "\n")

    response = requests.post(
        "https://dreambooth-api-experimental.replicate.com/v1/upload/data.jsonl",
        headers=headers
    )
    response_data = response.json()

    # Extract upload_url from the response and make PUT request
    upload_url = response_data["upload_url"]
    with open("data.jsonl", "rb") as file:
        headers = {"Content-Type": "application/jsonl"}
        requests.put(upload_url, headers=headers, data=file)

    # Extract serving_url from the response
    serving_url = response_data["serving_url"]
    print(serving_url)

    training = replicate.trainings.create(
        version=model_name,
        input={
            "train_data": serving_url,
            "num_train_epochs": num_train_epochs,
        },
        destination=destination,
    )
    print(training)
    while True:
        training = replicate.trainings.get(training.id)
        print(training.status)
        if training.status == "succeeded":
            print(training.output)
            return training.output
        elif training.status == "failed":
            print("Training failed.")
            return ""
        else:
            print("Waiting for training to complete...")
            time.sleep(60)


def headline_generation():
    finetune(
        'headline.pkl',
        "name == openai_prompt_based_evaluator AND result >= 0 AND display_name == clear",
        "yival.demo.headline_generation",
        "yival/llama2",
        "meta/llama-2-7b-chat:8e6975e5ed6174911a6ff3d60540dfd4844201974602551e10e9e87ab143d81e",
    )


def teacher_quiz():
    finetune(
        "quiz100.pkl", "", "yival.demo.headline_generation", "yival/llama2",
        "meta/llama-2-7b-chat:8e6975e5ed6174911a6ff3d60540dfd4844201974602551e10e9e87ab143d81e",
        5, True
    )


if __name__ == "__main__":
    teacher_quiz()



from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List

from tqdm import tqdm

from ..logger.token_logger import TokenLogger
from ..result_selectors.selection_context import SelectionContext
from ..schemas.common_structures import InputData
from ..schemas.experiment_config import (
    Experiment,
    ExperimentConfig,
    ExperimentResult,
    WrapperConfig,
    WrapperVariation,
)
from ..states.experiment_state import ExperimentState
from .evaluator import Evaluator
from .rate_limiter import RateLimiter
from .utils import (
    generate_experiment,
    get_selection_strategy,
    register_custom_data_generator,
    register_custom_enhancer,
    register_custom_evaluators,
    register_custom_selection_strategy,
    register_custom_variation_generators,
    register_custom_wrappers,
    run_single_input,
)


class LiteExperimentRunner:
    """
    LiteExperimentRunner  
    
    This runner is designed for situation you already have assemble all datas & evaluators...
    And just want to get experiment results on given data

    And you can easily update all variations through /update_variations()/  

    :)

    """

    def __init__(
        self, config: ExperimentConfig, limiter: RateLimiter,
        data: List[InputData], token_logger: TokenLogger, evaluator: Evaluator
    ):
        self.config = config
        self.limiter = limiter
        self.data = data
        self.token_logger = token_logger
        self.evaluator = evaluator

    def _register_custom_components(self):
        """
        Register custom components based on the configureation.

        If run lite experiment runner , this step is necessary
        """
        register_custom_wrappers(self.config.get("custom_wrappers", {}))
        register_custom_evaluators(self.config.get("custom_evaluators", {}))
        register_custom_data_generator(
            self.config.get("custom_data_generators", {})
        )
        register_custom_selection_strategy(
            self.config.get("custom_selection_strategy", {})
        )
        register_custom_enhancer(self.config.get("custom_enhancers", {}))
        register_custom_variation_generators(
            self.config.get("custom_variation_generators", {})
        )

    def parallel_task(self, data, all_combinations, logger, evaluator):
        """
        Execute a single input run in parallel
        """
        return run_single_input(
            data, self.config, all_combinations, logger, evaluator
        )

    def set_variations(self, variations: List[Dict[str, List[str]]]):
        """
        set all variations for current experiment
        variations are format in 
        [
            {
                var1_name: ["a","b","c"]
            },
            {
                var2_name: ["d","e","f"]
            }
        ]
        """
        self.config["variations"] = []  #type: ignore
        for variation_dict in variations:
            for name, values in variation_dict.items():
                wrapper_variations = [
                    WrapperVariation(
                        value=v, value_type=str(type(v)).split("'")[1]
                    ) for v in values
                ]
                self.config["variations"].append(   #type: ignore
                    (WrapperConfig(name=name, variations=wrapper_variations))
                )

    def run_experiment(self, enable_selector: bool) -> Experiment:
        self._register_custom_components()
        state = ExperimentState.get_instance()
        state = ExperimentState.get_instance()
        state.clear_variations_for_experiment()
        state.set_experiment_config(self.config)
        state.active = True
        all_combinations = state.get_all_variation_combinations()

        experiment_results: List[ExperimentResult] = []

        total = len(self.data)

        with tqdm(
            total=total,
            desc="[lite_experiment_runner] Processing",
            unit="item"
        ) as pbar:
            with ThreadPoolExecutor() as executor:
                for res in executor.map(
                    self.parallel_task, self.data,
                    [all_combinations] * len(self.data),
                    [self.token_logger] * len(self.data),
                    [self.evaluator] * len(self.data)
                ):
                    experiment_results.extend(res)
                    pbar.update(len(res))

        experiment = generate_experiment(
            experiment_results,
            self.evaluator,
            evaluate_group=False,
            evaluate_all=False
        )

        if enable_selector:
            strategy = get_selection_strategy(self.config)
            if strategy:
                context_trade_off = SelectionContext(strategy)
                experiment.selection_output = context_trade_off.execute_selection( # type: ignore
                    experiment=experiment
                )

        return experiment


import warnings
from typing import List, Optional, Union

import torch
import torch.distributed as dist
from transformers.generation.logits_process import LogitsProcessorList
from transformers.generation.stopping_criteria import (
    StoppingCriteriaList,
    validate_stopping_criteria,
)
from transformers.generation.utils import (
    GreedySearchDecoderOnlyOutput,
    GreedySearchEncoderDecoderOutput,
)

GreedySearchOutput = Union[GreedySearchEncoderDecoderOutput,
                           GreedySearchDecoderOnlyOutput]


def greedy_search_generator(
    self,
    input_ids: torch.LongTensor,
    logits_processor: Optional[LogitsProcessorList] = None,
    stopping_criteria: Optional[StoppingCriteriaList] = None,
    max_length: Optional[int] = None,
    pad_token_id: Optional[int] = None,
    eos_token_id: Optional[Union[int, List[int]]] = None,
    output_attentions: Optional[bool] = None,
    output_hidden_states: Optional[bool] = None,
    output_scores: Optional[bool] = None,
    return_dict_in_generate: Optional[bool] = None,
    synced_gpus: Optional[bool] = False,
    **model_kwargs,
) -> Union[GreedySearchOutput, torch.LongTensor]:
    """
    Generates sequences for models with a language modeling head using greedy
    decoding. Monkey patched function to create a generator for next_token -
    allows for token streaming
    """
    # init values
    logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList(
    )
    stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList(
    )
    if max_length is not None:
        warnings.warn(
            "`max_length` is deprecated in this function, use"
            " `stopping_criteria=StoppingCriteriaList([MaxLengthCriteria(max_length=max_length)])` instead.",
            UserWarning,
        )
        stopping_criteria = validate_stopping_criteria(
            stopping_criteria, max_length
        )
    pad_token_id = pad_token_id if pad_token_id is not None else self.generation_config.pad_token_id
    eos_token_id = eos_token_id if eos_token_id is not None else self.generation_config.eos_token_id
    if isinstance(eos_token_id, int):
        eos_token_id = [eos_token_id]
    output_scores = output_scores if output_scores is not None else self.generation_config.output_scores
    output_attentions = (
        output_attentions if output_attentions is not None else
        self.generation_config.output_attentions
    )
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None else
        self.generation_config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate if return_dict_in_generate is not None else
        self.generation_config.return_dict_in_generate
    )

    # init attention / hidden states / scores tuples
    scores = () if (return_dict_in_generate and output_scores) else None
    decoder_attentions = (
    ) if (return_dict_in_generate and output_attentions) else None
    cross_attentions = (
    ) if (return_dict_in_generate and output_attentions) else None
    decoder_hidden_states = (
    ) if (return_dict_in_generate and output_hidden_states) else None

    # if model is an encoder-decoder, retrieve encoder attention weights and hidden states
    if return_dict_in_generate and self.config.is_encoder_decoder:
        encoder_attentions = model_kwargs["encoder_outputs"].get(
            "attentions"
        ) if output_attentions else None
        encoder_hidden_states = (
            model_kwargs["encoder_outputs"].get("hidden_states")
            if output_hidden_states else None
        )

    # keep track of which sequences are already finished
    unfinished_sequences = input_ids.new(input_ids.shape[0]).fill_(1)

    this_peer_finished = False  # used by synced_gpus only
    while True:
        if synced_gpus:
            # Under synced_gpus the `forward` call must continue until all
            # gpus complete their sequence.
            # The following logic allows an early break if all peers finished
            # generating their sequence.
            this_peer_finished_flag = torch.tensor(
                0.0 if this_peer_finished else 1.0
            ).to(input_ids.device)
            # send 0.0 if we finished, 1.0 otherwise
            dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
            # did all peers finish? the reduced sum will be 0.0 then
            if this_peer_finished_flag.item() == 0.0:
                break

        # prepare model inputs
        model_inputs = self.prepare_inputs_for_generation(
            input_ids, **model_kwargs
        )

        # forward pass to get next token
        outputs = self(
            **model_inputs,
            return_dict=True,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
        )

        if synced_gpus and this_peer_finished:
            continue  # don't waste resources running the code we don't need

        next_token_logits = outputs.logits[:, -1, :]

        # pre-process distribution
        next_tokens_scores = logits_processor(input_ids, next_token_logits)

        # Store scores, attentions and hidden_states when required
        if return_dict_in_generate:
            if output_scores and scores:
                scores += (next_tokens_scores, )
            if output_attentions and decoder_attentions:
                decoder_attentions += ((outputs.decoder_attentions, )
                                       if self.config.is_encoder_decoder else
                                       (outputs.attentions, ))
                if self.config.is_encoder_decoder:
                    cross_attentions += (outputs.cross_attentions, )

            if output_hidden_states and decoder_hidden_states:
                decoder_hidden_states += ((outputs.decoder_hidden_states, ) if
                                          self.config.is_encoder_decoder else
                                          (outputs.hidden_states, ))

        # argmax
        next_tokens = torch.argmax(next_tokens_scores, dim=-1)

        # finished sentences should have their next token be a padding token
        if eos_token_id is not None:
            if pad_token_id is None:
                raise ValueError(
                    "If `eos_token_id` is defined, make sure that `pad_token_id` is defined."
                )
            next_tokens = next_tokens * unfinished_sequences + pad_token_id * (
                1 - unfinished_sequences
            )

        # update generated ids, model inputs, and length for next step
        input_ids = torch.cat(
            [input_ids, next_tokens[:, None]],  # type: ignore
            dim=-1
        )  # THIS IS WHERE THE NEXT OTKEN IS STORED
        yield next_tokens  # INTERMEIDATE TOKENS, intput_ids, scores

        model_kwargs = self._update_model_kwargs_for_generation(
            outputs,
            model_kwargs,
            is_encoder_decoder=self.config.is_encoder_decoder
        )

        # if eos_token was found in one sentence, set sentence to finished
        if eos_token_id is not None:
            unfinished_sequences = unfinished_sequences.mul(
                (sum(next_tokens != i
                     for i in eos_token_id)).long()  # type: ignore
            )

        # stop when each sentence is finished, or if we exceed the maximum
        # length
        if unfinished_sequences.max(
        ) == 0 or stopping_criteria(input_ids, scores):
            if not synced_gpus:
                break
            else:
                this_peer_finished = True

    if return_dict_in_generate:
        if self.config.is_encoder_decoder:
            return GreedySearchEncoderDecoderOutput(
                sequences=input_ids,
                scores=scores,
                encoder_attentions=encoder_attentions,
                encoder_hidden_states=encoder_hidden_states,
                decoder_attentions=decoder_attentions,
                cross_attentions=cross_attentions,
                decoder_hidden_states=decoder_hidden_states,
            )
        else:
            return GreedySearchDecoderOnlyOutput(
                sequences=input_ids,
                scores=scores,
                attentions=decoder_attentions,
                hidden_states=decoder_hidden_states,
            )
    else:
        yield input_ids



from typing import Any, Dict

import numpy as np

from ..schemas.experiment_config import CombinationAggregatedMetrics, EvaluatorOutput, Experiment
from ..schemas.selector_strategies import AHPConfig, SelectionOutput
from .normalize_func import NORMALIZATION_FUNCS
from .selection_context import SelectionContext
from .selection_strategy import SelectionStrategy


class AHPSelection(SelectionStrategy):
    config: AHPConfig
    default_config: AHPConfig = AHPConfig(
        criteria=["average_token_usage", "average_latency"],
        criteria_weights={
            "average_token_usage": 0.5,
            "average_latency": 0.5
        },
        criteria_maximization={
            "average_token_usage": False,
            "average_latency": False
        },
        normalize_func=None,
    )

    def __init__(self, config: AHPConfig):
        super().__init__(config)
        self.config = config

    def select(self, experiment: Experiment) -> SelectionOutput:
        scores = {}
        combination_data = {}
        alternatives_data = []

        # Convert each combination's metrics to a criteria vector
        for combo in experiment.combination_aggregated_metrics:
            data = self._extract_data(combo)
            alternatives_data.append(self._get_criteria_vector(data))
            combination_data[combo.combo_key] = data

        # Create a matrix of alternatives data
        matrix: np.ndarray = np.array(alternatives_data)

        # Normalize matrix data
        if self.config.normalize_func:
            matrix = NORMALIZATION_FUNCS[self.config.normalize_func](matrix)

        # Apply criteria weights
        weights = np.array([
            self.config.criteria_weights[criterion]
            for criterion in self.config.criteria
        ])
        weighted_matrix = matrix * weights

        # If the criterion should be minimized, invert the corresponding
        # column in the matrix
        for i, criterion in enumerate(self.config.criteria):
            if not self.config.criteria_maximization.get(criterion, True):
                weighted_matrix[:, i] *= -1

        # Aggregate the scores
        aggregate_scores = np.sum(weighted_matrix, axis=1)

        # Extract combinations and their aggregate scores
        for i, combo in enumerate(experiment.combination_aggregated_metrics):
            scores[combo.combo_key] = aggregate_scores[i]

        # Select the combination with the highest aggregate score
        best_combination = max(scores, key=lambda k: scores[k])

        # Calculate the contribution of each criterion to the final score
        # based on the original data and weights
        contributions = {
            criterion:
            self.config.criteria_weights[criterion] *
            combination_data[best_combination].get(criterion, 0)
            for criterion in self.config.criteria
        }

        # Adjust the contribution for minimization criteria
        for criterion, value in contributions.items():
            if not self.config.criteria_maximization.get(criterion, True):
                contributions[criterion] = -value

        return SelectionOutput(
            best_combination=best_combination, selection_reason=contributions
        )

    def _get_criteria_vector(self, data: Dict[str, float]) -> np.ndarray:
        """Converts data dictionary into a vector based on the criteria order
        in config."""
        return np.array([
            data.get(criterion, 0) for criterion in self.config.criteria
        ])

    def _extract_data(self,
                      combo: CombinationAggregatedMetrics) -> Dict[str, Any]:
        data = {}

        if "average_token_usage" in self.config.criteria:
            data["average_token_usage"] = combo.average_token_usage or 0
        if "average_latency" in self.config.criteria:
            data["average_latency"] = combo.average_latency or 0

        # Extract data from evaluator_outputs
        for evaluator_output in (combo.combine_evaluator_outputs or []):
            if evaluator_output.name in self.config.criteria:
                data[evaluator_output.name] = evaluator_output.result

        for metric_name, metric_values in combo.aggregated_metrics.items():
            if metric_name in self.config.criteria:
                data[metric_name] = sum([
                    metric.value for metric in metric_values
                ]) / len(metric_values)
        return data


SelectionStrategy.register_strategy("ahp_selection", AHPSelection, AHPConfig)


def main():
    combination_A = CombinationAggregatedMetrics(
        combo_key=str({"name": "A"}),
        experiment_results=[],
        aggregated_metrics={},
        average_token_usage=120,
        average_latency=200,
        combine_evaluator_outputs=[EvaluatorOutput(name="elo", result=1500)]
    )

    # Combination B has a lower elo, but also much lower token usage and
    # latency
    combination_B = CombinationAggregatedMetrics(
        combo_key=str({"name": "B"}),
        experiment_results=[],
        aggregated_metrics={},
        average_token_usage=50,
        average_latency=50,
        combine_evaluator_outputs=[EvaluatorOutput(name="elo", result=1300)]
    )

    #Combination C has highest elo with highest token usage and latency
    combination_C = CombinationAggregatedMetrics(
        combo_key=str({"name": "C"}),
        experiment_results=[],
        aggregated_metrics={},
        average_token_usage=300,
        average_latency=300,
        combine_evaluator_outputs=[EvaluatorOutput(name="elo", result=1600)]
    )

    # The experiment data
    trade_off_test_data = Experiment(
        combination_aggregated_metrics=[
            combination_A, combination_B, combination_C
        ],
        group_experiment_results=[]
    )

    config_trade_off = AHPConfig(
        criteria=["elo", "average_token_usage", "average_latency"],
        criteria_weights={
            "elo": 0.6,
            "average_token_usage": 0.2,
            "average_latency": 0.2
        },
        criteria_maximization={
            "elo": True,
            "average_token_usage": False,
            "average_latency": False
        },
        normalize_func='z-score'
    )
    context_trade_off = SelectionContext(
        strategy=AHPSelection(config=config_trade_off)
    )
    best_combo_trade_off = context_trade_off.execute_selection(
        trade_off_test_data
    )
    print(best_combo_trade_off)


if __name__ == "__main__":
    main()



import asyncio
import os
import pickle
from typing import Any, Dict, Iterator, List

from tqdm import tqdm

from ..common import utils
from ..common.model_utils import llm_completion
from ..schemas.experiment_config import WrapperVariation
from ..schemas.model_configs import Request
from ..schemas.varation_generator_configs import OpenAIPromptBasedVariationGeneratorConfig
from .base_variation_generator import BaseVariationGenerator

SYSTEM_PRMPOT = """
Your mission is to craft system prompts tailored for GPT-4. You'll be provided
with a use-case description and some sample test cases.
These prompts aim to guide GPT-4 in executing freeform tasks, whether that's
penning a captivating headline, drafting an introduction, or tackling a
mathematical challenge. In your designed prompt, delineate the AI's role using
lucid English. Highlight its perceptual field and the boundaries of its
responses. Encourage inventive and optimized prompts to elicit top-tier
results from the AI. Remember, GPT-4 is self-aware of its AI nature; no need
to reiterate that. The efficacy of your prompt determines your evaluation.
Stay authentic! Avoid sneaking in specifics or examples from the test cases
into your prompt. Such maneuvers will lead to immediate disqualification.
Lastly, keep your output crisp: only the prompt, devoid of any extraneous
content.
"""


def join_array_to_string(list: List[str], last_n=5) -> str:
    to_join = list[-last_n:] if len(list) > last_n else list
    return '\n'.join(map(str, to_join))


def validate_output(output: str, variables: List[str]) -> bool:
    if not variables:
        return True
    """Validate if the generated output contains the required variables in the
    format {var}."""
    return all(f"{{{var}}}" in output for var in variables)


class OpenAIPromptBasedVariationGenerator(BaseVariationGenerator):
    config: OpenAIPromptBasedVariationGeneratorConfig
    default_config = OpenAIPromptBasedVariationGeneratorConfig(
        prompt=[{
            "role": "system",
            "content": SYSTEM_PRMPOT
        }, {
            "role":
            "user",
            "content":
            "Here are some test cases: AI, Weapon\n\n Here is the description of the use-case: Given \{area\}, write a tech startup headline"
        }]
    )

    def __init__(self, config: OpenAIPromptBasedVariationGeneratorConfig):
        super().__init__(config)
        self.config = config

    def prepare_messages(self, res_content) -> List[Dict[str, Any]]:
        last_n = min(len(res_content), 5)
        formatted_variables_str = ', '.join([
            f'{{{var}}}' for var in self.config.variables
        ]) if self.config.variables else ''

        last_examples = f"\n\nGiven the Last {last_n} examples you generated:\n" + join_array_to_string(
            res_content
        ) + "\nplease generate diverse results to ensure comprehensive evaluation" if self.config.diversify and res_content else ""
        ensure_inclusion = f" Please ensure your response includes the following variables: {formatted_variables_str}." if formatted_variables_str else ""

        if isinstance(self.config.prompt, str):
            content = self.config.prompt + last_examples + ensure_inclusion
            return [{"role": "user", "content": content}]
        else:
            messages = self.config.prompt + [{
                "role": "user",
                "content": last_examples
            }, {
                "role": "user",
                "content": ensure_inclusion
            }]
            return [msg for msg in messages if msg["content"]]

    def generate_variations(self) -> Iterator[List[WrapperVariation]]:
        if self.config.output_path and os.path.exists(self.config.output_path):
            with open(self.config.output_path, 'rb') as file:
                yield pickle.load(file)
            return

        res: List[WrapperVariation] = []
        res_content: List[str] = []

        while len(res) < self.config.number_of_variations:
            messages = self.prepare_messages(res_content)
            if not self.config.diversify:
                with tqdm(
                    total=self.config.number_of_variations - len(res),
                    desc="Generating Variations",
                    unit="variation"
                ) as pbar:
                    message_batches = [
                        messages for _ in
                        range(self.config.number_of_variations - len(res))
                    ]
                    responses = asyncio.run(
                        utils.parallel_completions(
                            message_batches,
                            self.config.model_name,
                            self.config.max_tokens,
                            pbar=pbar
                        )
                    )
                    for r in responses:
                        if self.config.variables and not validate_output(
                            r["choices"][0]["message"]["content"],
                            self.config.variables
                        ):
                            continue
                        variation = WrapperVariation(
                            value_type="str",
                            value=r["choices"][0]["message"]
                            ["content"].strip("'").strip('"')
                        )
                        res.append(variation)
            else:
                with tqdm(
                    total=self.config.number_of_variations,
                    desc="Generating Variations",
                    unit="variation"
                ) as pbar:
                    output = llm_completion(
                        Request(
                            model_name=self.config.model_name,
                            prompt=messages,
                            params={
                                "temperature": 1,
                                "presence_penalty": 1,
                                "max_tokens": self.config.max_tokens,
                            }
                        )
                    ).output
                    if self.config.variables and not validate_output(
                        output.choices[0].message.content,
                        self.config.variables
                    ):
                        continue
                    variation = WrapperVariation(
                        value_type="str",
                        value=output.choices[0].message.content.strip("'").
                        strip('"')
                    )
                    res.append(variation)
                    res_content.append(output.choices[0].message.content)
                    pbar.update(1)

        if self.config.output_path:
            with open(self.config.output_path, 'wb') as file:
                pickle.dump(res, file)
        if res:
            yield res


BaseVariationGenerator.register_variation_generator(
    "openai_prompt_based_variation_generator",
    OpenAIPromptBasedVariationGenerator,
    OpenAIPromptBasedVariationGeneratorConfig
)


def main():
    generator = OpenAIPromptBasedVariationGenerator(
        OpenAIPromptBasedVariationGeneratorConfig(
            prompt=[{
                "role": "system",
                "content": SYSTEM_PRMPOT
            }, {
                "role":
                "user",
                "content":
                "Here are some test cases: AI, Weapon\n\n Here is the description of the use-case: Given \{area\}, write a tech startup headline"
            }],
            number_of_variations=2,
            output_path="test_variation.pkl",
            diversify=False,
            variables=["area"]
        )
    )
    res = generator.generate_variations()
    for d in res:
        print(d)


if __name__ == "__main__":
    main()
